import json
import re
import shutil
from enum import Enum
from pathlib import Path
from typing import List, Optional, Union

import nibabel as nib
import pandas as pd
from nilearn.image import concat_imgs

from clinica.iotools.bids_utils import StudyName, bids_id_factory
from clinica.utils.stream import cprint, log_and_raise

__all__ = [
    "read_clinical_data",
    "define_participants",
    "write_subject_data",
    "write_sessions",
    "write_scans",
    "write_participants",
    "check_modalities",
]


def _get_subjects_list_from_data(data_directory: Path) -> List[str]:
    return list(
        dict.fromkeys(
            re.match(r"IXI\d{3}", path.name).group(0)
            for path in data_directory.rglob(pattern="IXI*.nii.gz")
            if re.match(r"IXI\d{3}", path.name)
        )
    )


def _get_subjects_list_from_file(subjs_list_path: Path) -> List[str]:
    return subjs_list_path.read_text().splitlines()


def define_participants(
    data_directory: Path,
    subjs_list_path: Optional[Path] = None,
) -> List[str]:
    """
    Defines the actual list of participants based on the (provided) subjects list filtered
    using existing data.

    Parameters
    ----------
    data_directory : Path to the raw data directory.
    subjs_list_path : [Optional] Path to a text file containing a list of specific subjects to extract.

    Returns
    -------
    The list of ids that were either present in the data directory or asked for with a specific text file, provided
    associated images actually exists.

    """

    list_from_data = _get_subjects_list_from_data(data_directory)
    if subjs_list_path is None:
        return sorted(list_from_data)
    cprint("Loading a subjects list provided by the user...")
    list_from_file = _get_subjects_list_from_file(subjs_list_path)
    list_filtered = [subject for subject in list_from_file if subject in list_from_data]
    invalid_subjects = list(set(list_from_file) - set(list_filtered))
    if invalid_subjects:
        cprint(
            f"The subjects : {' , '.join(invalid_subjects)} do not have any associated data inside the directory {data_directory}"
            f" and can not be converted."
        )
    return sorted(list_filtered)


def _rename_clinical_data_to_bids(column: str) -> str:
    if column == "SEX_ID (1=m, 2=f)":
        return "sex"
    if column == "ETHNIC_ID":
        return "ethnicity"
    if column == "MARITAL_ID":
        return "marital status"
    if column == "OCCUPATION_ID":
        return "occupation"
    if column == "QUALIFICATION_ID":
        return "qualification"
    if column == "IXI_ID":
        return "source_id"
    if column == "DOB":
        return "date of birth"
    if column == "STUDY_DATE":
        return "acq_time"
    else:
        return column.lower()


def _get_sex_mapping() -> pd.DataFrame:
    return pd.DataFrame({"SEX": ["male", "female"]}, index=[1, 2])["SEX"]


class ClinicalDataMapping(str, Enum):
    ETHNIC = "Ethnicity"
    MARITAL = "Marital Status"
    OCCUPATION = "Occupation"
    QUALIFICATION = "Qualification"


def _get_mapping(clinical_data_path: Path, map: ClinicalDataMapping) -> pd.Series:
    try:
        return pd.read_excel(
            clinical_data_path / "IXI.xls", sheet_name=map.value
        ).set_index("ID")[map.name]
    except FileNotFoundError:
        log_and_raise(
            f"Clinical data stored in the folder {clinical_data_path} is expected to be an excel file named 'IXI.xls'. "
            f"In case the file downloaded from the IXI website changed format, please do not hesitate to report to us !",
            FileNotFoundError,
        )
    except (ValueError, KeyError):
        log_and_raise(
            f"{map.value} mapping is expected to be contained in a sheet called {map.value} coming from the clinical data excel. "
            f"Possibilities are supposed to be described in a {map.name} column associated to keys from the 'ID' column. "
            f"In case the file downloaded from the IXI website changed format, please do not hesitate to report to us !",
            ValueError,
        )


def _padding_source_id(source_id: Union[str, int]) -> str:
    if len(str(source_id)) > 3:
        log_and_raise(
            f"The source id {source_id} has more than 3 digits while IXI"
            f"source ids are expected to be between 1 and 3 digits.",
            ValueError,
        )
    return f"IXI{'0'* (3 - len(str(source_id))) + str(source_id)}"


def read_clinical_data(clinical_data_path: Path) -> pd.DataFrame:
    """
    Reads and formats IXI clinical data.

    Parameters
    ----------
    clinical_data_path : Path to the directory where the clinical data .xls is stored.

    Returns
    -------
    A dataframe containing the clinical data, with some modifications: padded study id and added session id.

    """
    try:
        clinical_data = pd.read_excel(clinical_data_path / "IXI.xls")
    except FileNotFoundError:
        log_and_raise(
            f"Clinical data stored in the folder {clinical_data_path} is expected to be an excel file named 'IXI.xls'. "
            f"In case the file downloaded from the IXI website changed format, please do not hesitate to report to us !",
            FileNotFoundError,
        )
    else:
        if "DATE_AVAILABLE" in clinical_data.columns:
            clinical_data.drop("DATE_AVAILABLE", axis=1, inplace=True)
        clinical_data["SEX_ID (1=m, 2=f)"] = clinical_data["SEX_ID (1=m, 2=f)"].map(
            _get_sex_mapping()
        )
        for mapping in ClinicalDataMapping:
            clinical_data[f"{mapping.name}_ID"] = clinical_data[
                f"{mapping.name}_ID"
            ].map(_get_mapping(clinical_data_path, mapping))

        clinical_data["IXI_ID"] = clinical_data.IXI_ID.apply(
            lambda x: _padding_source_id(x)
        )
        clinical_data.rename(
            lambda x: _rename_clinical_data_to_bids(x), axis=1, inplace=True
        )
        clinical_data.fillna("n/a", inplace=True)
        clinical_data["session_id"] = "ses-M000"
        return clinical_data


def _rename_modalities(input_mod: str) -> str:
    if input_mod == "T1":
        return "T1w"
    if input_mod == "T2":
        return "T2w"
    if input_mod == "MRA":
        return "angio"
    if input_mod == "PD":
        return "PDw"
    if input_mod == "DTI":
        return "dti"
    raise ValueError(f"The modality {input_mod} is not recognized in the IXI dataset.")


def _define_magnetic_field(hospital: str) -> str:
    if hospital in ("Guys", "IOP"):
        return "1.5"
    if hospital == "HH":
        return "3"
    raise ValueError(f"The hospital {hospital} was not recognized.")


def _get_img_data(data_directory: Path) -> pd.DataFrame:
    """Finds paths for all images that are not DTI data and processes the info contained in their names"""
    df = pd.DataFrame(
        {
            "img_path": [
                path
                for path in data_directory.rglob(pattern="IXI*.nii.gz")
                if re.search(r"IXI\d{3}(-\w*){3}.nii.gz$", str(path))
            ]
        }
    )
    df = (
        df.assign(img_name=lambda df: df.img_path.apply(lambda x: x.name))
        .assign(img_name_no_ext=lambda df: df.img_name.apply(lambda x: x.split(".")[0]))
        .assign(subject=lambda df: df.img_name_no_ext.apply(lambda x: x.split("-")[0]))
        .assign(
            participant_id=lambda df: df.subject.apply(
                lambda x: bids_id_factory(StudyName.IXI).from_original_study_id(x)
            )
        )
        .assign(hospital=lambda df: df.img_name_no_ext.apply(lambda x: x.split("-")[1]))
        .assign(
            modality=lambda df: df.img_name_no_ext.apply(
                lambda x: _rename_modalities(x.split("-")[3])
            )
        )
        .assign(field=lambda df: df.hospital.apply(lambda x: _define_magnetic_field(x)))
        .assign(session="ses-M000")
    )
    return df


def _get_bids_filename_from_image_data(img: pd.Series) -> str:
    return f"{img['participant_id']}_{img['session']}_{img['modality']}"


def write_subject_data(bids_dir: Path, participant: str, path_to_dataset: Path) -> None:
    """
    Writes the data of the IXI subject in the BIDS directory following BIDS specifications.

    Parameters
    ----------
    bids_dir : Path to the output BIDS directory.
    participant : Current converted subject study id (str).
    path_to_dataset : Path to the raw dataset directory.
    """
    data_df = _get_img_data(path_to_dataset)
    _write_subject_no_dti(data_df[data_df["subject"] == participant], bids_dir)
    _write_subject_dti_if_exists(bids_dir, participant, path_to_dataset)


def _write_json_image(writing_path: Path, hospital: str, field: str) -> None:
    """
    Writes a json associated to one IXI image.

    Parameters
    ----------
    writing_path : Path indicating under what name to write the json.
    hospital : identifier for the hospital responsible for the acquisition, determined from the image filename.
    field :  magnetic field used for the acquisition of the image, determined from the hospital.
    """
    with open(writing_path, "w") as f:
        json.dump(
            {
                "InstitutionName": hospital,
                "MagneticFieldStrength (T)": field,
            },
            f,
            indent=4,
        )


def _write_subject_no_dti(subject_df: pd.DataFrame, bids_path: Path) -> None:
    """Copies all subject data but DTI"""
    for _, row in subject_df.iterrows():
        cprint(
            f"Converting modality {row['modality']} for subject {row['subject']}.",
            lvl="debug",
        )
        filename = _get_bids_filename_from_image_data(row)
        data_path = bids_path / row["participant_id"] / row["session"] / "anat"
        data_path.mkdir(parents=True, exist_ok=True)
        shutil.copy2(row["img_path"], f"{data_path}/{filename}.nii.gz")
        _write_json_image(
            (data_path / filename).with_suffix(".json"), row["hospital"], row["field"]
        )


def _write_subject_dti_if_exists(
    bids_path: Path, subject: str, data_directory: Path
) -> None:
    """Processes DTI data if found for a subject"""
    if dti_paths := _find_subject_dti_data(data_directory, subject):
        cprint(f"Converting modality DTI for subject {subject}.", lvl="debug")
        dti_to_save = _merge_dti(dti_paths)
        bids_id = bids_id_factory(StudyName.IXI).from_original_study_id(subject)
        data_path = bids_path / bids_id / "ses-M000" / "dwi"
        data_path.mkdir(parents=True, exist_ok=True)
        filename = Path(f"{bids_id}_ses-M000_dwi")
        dti_to_save.to_filename(data_path / filename.with_suffix(".nii.gz"))
        hospital = dti_paths[0].name.split("-")[1]
        _write_json_image(
            data_path / filename.with_suffix(".json"),
            hospital,
            _define_magnetic_field(hospital),
        )


def _find_subject_dti_data(data_directory: Path, subject: str) -> List[Path]:
    pattern = subject + r"(-\w*){2}-DTI(-\w*){1}.nii.gz$"
    return [
        path
        for path in data_directory.rglob(pattern="IXI*.nii.gz")
        if re.search(pattern, str(path))
    ]


def _merge_dti(dti_images: List[Path]) -> nib.Nifti1Image:
    return concat_imgs([nib.load(img) for img in dti_images])


def write_scans(bids_dir: Path, participant: str) -> None:
    """
    Write the scans.tsv for the only session (ses-M000) of a IXI subject.

    Parameters
    ----------
    bids_dir : Path to the output BIDS directory.
    participant : Current converted subject study id (str).
    """
    bids_id = bids_id_factory(StudyName.IXI).from_original_study_id(participant)
    to_write = pd.DataFrame(
        {
            "filename": [
                f"{path.parent.name}/{path.name}"
                for path in (bids_dir / bids_id).rglob(f"{bids_id}*.nii.gz")
            ]
        }
    )
    to_write.to_csv(
        bids_dir / bids_id / "ses-M000" / f"{bids_id}_ses-M000_scans.tsv",
        sep="\t",
        index=False,
    )


def write_sessions(
    bids_dir: Path, clinical_data: pd.DataFrame, participant: str
) -> None:
    """
    Writes the sessions.tsv for a IXI subject.

    Parameters
    ----------
    bids_dir : Path to the output BIDS directory.
    clinical_data : Dataframe containing the formatted clinical data of the IXI study.
    participant : Current converted subject study id (str).
    """
    line = clinical_data[clinical_data["source_id"] == participant]
    bids_id = bids_id_factory(StudyName.IXI).from_original_study_id(participant)
    line[["source_id", "session_id", "acq_time"]].to_csv(
        bids_dir / bids_id / f"{bids_id}_sessions.tsv", sep="\t", index=False
    )


def write_participants(
    bids_dir: Path, clinical_data: pd.DataFrame, participants: List[str]
) -> None:
    """
    Write the participants.tsv at the root of the BIDS directory.

    Parameters
    ----------
    bids_dir : Path to the output BIDS directory.
    clinical_data : Dataframe containing the formatted clinical data of the IXI study.
    participants : List of converted subjects study source ids.
    """
    clinical_data.set_index("source_id", inplace=True, drop=False)
    clinical_data.assign(
        participant_id=clinical_data.source_id.apply(
            lambda x: bids_id_factory(StudyName.IXI).from_original_study_id(x)
        )
    )
    for participant in participants:
        if participant not in clinical_data.index:
            clinical_data.loc[participant] = "n/a"
            clinical_data.loc[participant, "source_id"] = participant
    if not bids_dir.exists():
        bids_dir.mkdir()
    clinical_data.loc[participants].drop(["acq_time", "session_id"], axis=1).to_csv(
        bids_dir / "participants.tsv", sep="\t", index=False, na_rep="n/a"
    )
    clinical_data.reset_index(drop=True, inplace=True)


def _identify_expected_modalities(data_directory: Path) -> List[str]:
    return [
        p.name.split("-")[1]
        for p in data_directory.iterdir()
        if p.is_dir() and "IXI-" in str(p)
    ]


def check_modalities(data_directory: Path, participants: List[str]) -> None:
    """
    Verify what modality folders are available in the given data directory and checks if some are missing per participant

    Parameters
    ----------
    data_directory : Path to raw dataset.
    participants : List of the subject ids of all participants

    """
    expected_modalities = sorted(_identify_expected_modalities(data_directory))
    message = (
        f"Modalities : {' , '.join(_rename_modalities(mod) for mod in expected_modalities)}"
        f" were identified inside {data_directory} for conversion.\n"
    )
    participants_missing_mod = dict()
    for participant in participants:
        missing_mods = []
        for mod in expected_modalities:
            if not list(data_directory.rglob(f"{participant}*{mod}*.nii.gz")):
                missing_mods += [_rename_modalities(mod)]
        if missing_mods:
            participants_missing_mod[participant] = missing_mods
    if participants_missing_mod:
        message += f"Some subjects do not have data for the following modalities :\n"
        for sub, mod in participants_missing_mod.items():
            message += f"{sub} : {' , '.join(mod)}\n"

    cprint(message)



from dataclasses import dataclass
from os import PathLike
from pathlib import Path
from typing import List

import pandas as pd

__all__ = [
    "compute_missing_mods",
    "compute_missing_processing",
]


@dataclass
class SubjectSession:
    participant_id: str
    session_id: str
    session_path: Path

    @property
    def sub_ses_id(self) -> str:
        return f"{self.participant_id}_{self.session_id}"


def compute_missing_mods(
    bids_dir: PathLike, out_dir: PathLike, output_prefix: str = ""
) -> None:
    """Compute the list of missing modalities for each subject in a BIDS compliant dataset.

    Parameters
    ----------
    bids_dir : PathLike
        Path to the BIDS directory.

    out_dir : PathLike
        Path to the output folder.

    output_prefix : str, optional
        String that replaces the default prefix ('missing_mods_')
        in the name of all the created output files.
        Default = "".
    """
    import os
    from glob import glob
    from os import path
    from pathlib import Path

    import pandas as pd

    from clinica.iotools.converter_utils import (
        MissingModsTracker,
        write_longitudinal_analysis,
        write_statistics,
    )

    out_dir = Path(out_dir)
    bids_dir = Path(bids_dir)
    out_dir.mkdir(parents=True, exist_ok=True)

    # Find all the modalities and sessions available for the input dataset
    mods_and_sess = _find_mods_and_sess(bids_dir)
    sessions_found = mods_and_sess["sessions"]
    mods_and_sess.pop("sessions")
    mods_avail_dict = mods_and_sess
    mods_avail = [j for i in mods_avail_dict.values() for j in i]
    cols_dataframe = mods_avail[:]
    cols_dataframe.insert(0, "participant_id")
    mmt = MissingModsTracker(sessions_found, mods_avail)

    out_file_name = "missing_mods_" if output_prefix == "" else output_prefix + "_"

    missing_mods_df = pd.DataFrame(columns=cols_dataframe)
    row_to_append_df = pd.DataFrame(columns=cols_dataframe)
    subjects_paths_lists = glob(path.join(bids_dir, "*sub-*"))
    subjects_paths_lists.sort()

    if len(subjects_paths_lists) == 0:
        raise IOError("No subjects found or dataset not BIDS compliant.")
    # Check the modalities available for each session
    for ses in sessions_found:
        for sub_path in subjects_paths_lists:
            mods_avail_bids = []
            subj_id = sub_path.split(os.sep)[-1]
            row_to_append_df["participant_id"] = pd.Series(subj_id)
            ses_path_avail = glob(path.join(sub_path, ses))
            if len(ses_path_avail) == 0:
                mmt.increase_missing_ses(ses)
                for mod in mods_avail:
                    row_to_append_df[mod] = pd.Series("0")
            else:
                ses_path = ses_path_avail[0]
                mods_paths_folders = glob(path.join(ses_path, "*/"))

                for p in mods_paths_folders:
                    p = p[:-1]
                    mods_avail_bids.append(p.split("/").pop())

                # Check if a modality folder is available and if is empty
                if "func" in mods_avail_bids:
                    # Extract all the task available
                    for m in mods_avail_dict["func"]:
                        tokens = m.split("_")
                        task_name = tokens[1]
                        task_avail_list = glob(
                            path.join(ses_path, "func", "*" + task_name + "*")
                        )

                        if len(task_avail_list) == 0:
                            row_to_append_df[m] = pd.Series("0")
                        else:
                            row_to_append_df[m] = pd.Series("1")
                # If the folder is not available but the modality is
                # in the list of the available one mark it as missing
                else:
                    if "func" in mods_avail_dict:
                        for m in mods_avail_dict["func"]:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)

                if "dwi" in mods_avail_bids:
                    row_to_append_df["dwi"] = pd.Series("1")
                else:
                    if "dwi" in mods_avail:
                        row_to_append_df["dwi"] = pd.Series("0")
                        mmt.add_missing_mod(ses, "dwi")

                if "anat" in mods_avail_bids:
                    for m in mods_avail_dict["anat"]:
                        anat_aval_list = glob(path.join(ses_path, "anat", "*.nii.gz"))
                        anat_aval_list = [
                            elem for elem in anat_aval_list if m.lower() in elem.lower()
                        ]
                        if len(anat_aval_list) > 0:
                            row_to_append_df[m] = pd.Series("1")
                        else:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)
                else:
                    if "anat" in mods_avail_dict:
                        for m in mods_avail_dict["anat"]:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)

                if "fmap" in mods_avail_bids:
                    row_to_append_df["fmap"] = pd.Series("1")
                else:
                    if "fmap" in mods_avail:
                        row_to_append_df["fmap"] = pd.Series("0")
                        mmt.add_missing_mod(ses, "fmap")
                if "pet" in mods_avail_bids:
                    # Extract all the task available
                    for m in mods_avail_dict["pet"]:
                        tokens = m.split("_")
                        pet_acq = tokens[1]
                        acq_avail_list = glob(
                            path.join(ses_path, "pet", "*" + pet_acq + "*")
                        )

                        if len(acq_avail_list) == 0:
                            row_to_append_df[m] = pd.Series("0")
                        else:
                            row_to_append_df[m] = pd.Series("1")
                # If the folder is not available but the modality is
                # in the list of the available one mark it as missing
                else:
                    if "pet" in mods_avail_dict:
                        for m in mods_avail_dict["pet"]:
                            row_to_append_df[m] = pd.Series("0")
                            mmt.add_missing_mod(ses, m)

            missing_mods_df = pd.concat([missing_mods_df, row_to_append_df])
            row_to_append_df = pd.DataFrame(columns=cols_dataframe)

        missing_mods_df = missing_mods_df[cols_dataframe]
        missing_mods_df.to_csv(
            path.join(out_dir, out_file_name + ses + ".tsv"),
            sep="\t",
            index=False,
            encoding="utf-8",
        )
        missing_mods_df = pd.DataFrame(columns=cols_dataframe)

    write_statistics(
        out_dir / (out_file_name + "summary.txt"),
        len(subjects_paths_lists),
        sessions_found,
        mmt,
    )
    write_longitudinal_analysis(
        out_dir / "analysis.txt", bids_dir, out_dir, sessions_found, out_file_name
    )


def compute_missing_processing(
    bids_dir: PathLike, caps_dir: PathLike, out_file: PathLike
):
    """Compute the list of missing processing for each subject in a CAPS compliant dataset.

    Parameters
    ----------
    bids_dir : PathLike
        Path to the BIDS directory.

    caps_dir : PathLike
        Path to the CAPS directory.

    out_file : PathLike
        Path to the output file (filename included).
    """
    bids_dir = Path(bids_dir)
    caps_dir = Path(caps_dir)
    output_df = pd.DataFrame()
    groups = _get_groups(caps_dir)
    tracers = _get_pet_tracers(bids_dir)

    for subject_path in (caps_dir / "subjects").glob("sub-*"):
        participant_id = subject_path.parent.name
        for session_path in subject_path.glob("ses-*"):
            session_id = session_path.parent.name
            subject = SubjectSession(participant_id, session_id, session_path)
            row = _compute_missing_processing_single_row(subject, groups, tracers)
            output_df = pd.concat([output_df, row])
    output_df.sort_values(["participant_id", "session_id"], inplace=True)
    output_df.to_csv(out_file, sep="\t", index=False)


def _get_groups(caps_dir: Path) -> List[str]:
    if (caps_dir / "groups").exists():
        return [f.name for f in (caps_dir / "groups").iterdir()]
    return []


def _get_pet_tracers(bids_dir: Path) -> List[str]:
    """Retrieve pet tracers available."""
    from clinica.utils.stream import cprint

    modalities = _find_mods_and_sess(bids_dir)
    modalities.pop("sessions")
    tracers = [
        j.split("_")[1]
        for modality in modalities.values()
        for j in modality
        if "pet" in j
    ]
    cprint(f"Available tracers: {tracers}", lvl="info")

    return tracers


def _compute_missing_processing_single_row(
    subject: SubjectSession,
    groups: List[str],
    tracers: List[str],
) -> pd.DataFrame:
    """Compute a single row of the missing processing dataframe."""
    row = pd.DataFrame(
        [[subject.participant_id, subject.session_id]],
        columns=["participant_id", "session_id"],
    )
    row = _compute_missing_processing_t1_volume(row, groups, subject)
    for pipeline in ("t1-linear", "flair_linear"):
        row[0, pipeline] = "1" if (subject.session_path / pipeline).exists() else "0"
    row[0, "t1-freesurfer"] = (
        "1"
        if (subject.session_path / "t1" / "freesurfer_cross_sectional").exists()
        else "0"
    )
    row = _compute_missing_processing_pet_volume(
        row, groups, tracers, subject.session_path
    )
    for tracer in tracers:
        row.loc[0, f"pet-surface_{tracer}"] = (
            "1"
            if len(
                [
                    f
                    for f in (subject.session_path / "pet" / "surface").glob(
                        f"*{tracer}*"
                    )
                ]
            )
            > 0
            else "0"
        )
    for tracer in tracers:
        row.loc[0, f"pet-linear_{tracer}"] = (
            "1"
            if len(
                [f for f in (subject.session_path / "pet_linear").glob(f"*{tracer}*")]
            )
            > 0
            else "0"
        )
    return row


def _compute_missing_processing_t1_volume(
    df: pd.DataFrame,
    groups: List[str],
    subject: SubjectSession,
) -> pd.DataFrame:
    if (subject.session_path / "t1" / "spm" / "segmentation").exists():
        df.loc[0, "t1-volume-segmentation"] = "1"
        for group in groups:
            group_id = group.split("-")[-1]
            filename = f"{subject.sub_ses_id}_T1w_target-{group_id}_transformation-forward_deformation.nii.gz"
            if (
                subject.session_path / "t1" / "spm" / "dartel" / group / filename
            ).exists():
                df.loc[0, f"t1-volume-register-dartel_{group}"] = "1"
                pattern = f"{subject.sub_ses_id}_T1w_segm-*_space-Ixi549Space_modulated-*_probability.nii.gz"
                dartel2mni = [
                    f
                    for f in (
                        subject.session_path / "t1" / "spm" / "dartel" / group
                    ).glob(pattern)
                ]
                if len(dartel2mni) > 0:
                    df.loc[0, f"t1-volume-dartel2mni_{group}"] = "1"
                    if (
                        subject.session_path
                        / "t1"
                        / "spm"
                        / "dartel"
                        / group
                        / "atlas_statistics"
                    ).exists():
                        df.loc[0, f"t1-volume-parcellation_{group}"] = "1"
                    else:
                        df.loc[0, f"t1-volume-parcellation_{group}"] = "0"
                else:
                    df.loc[0, f"t1-volume-dartel2mni_{group}"] = "0"
                    df.loc[0, f"t1-volume-parcellation_{group}"] = "0"
            else:
                df.loc[0, f"t1-volume-register-dartel_{group}"] = "0"
                df.loc[0, f"t1-volume-dartel2mni_{group}"] = "0"
                df.loc[0, f"t1-volume-parcellation_{group}"] = "0"
    else:
        df.loc[0, "t1-volume-segmentation"] = "0"
        for group in groups:
            df.loc[0, f"t1-volume-register-dartel_{group}"] = "0"
            df.loc[0, f"t1-volume-dartel2mni_{group}"] = "0"
            df.loc[0, f"t1-volume-parcellation_{group}"] = "0"
    return df


def _compute_missing_processing_pet_volume(
    df: pd.DataFrame,
    groups: List[str],
    tracers: List[str],
    session_path: Path,
) -> pd.DataFrame:
    for group in groups:
        for tracer in tracers:
            pet_paths_pvc = [
                f
                for f in (session_path / "pet" / "preprocessing" / group).glob(
                    f"*{tracer}*"
                )
                if "pvc" in f
            ]
            pet_paths_no_pvc = [
                f
                for f in (session_path / "pet" / "preprocessing" / group).glob(
                    f"*{tracer}*"
                )
                if "pvc" not in f
            ]
            df.loc[0, f"pet-volume_{tracer}_{group}_pvc-True"] = (
                "1" if len(pet_paths_pvc) > 0 else "0"
            )
            df.loc[0, f"pet-volume_{tracer}_{group}_pvc-False"] = (
                "1" if len(pet_paths_no_pvc) > 0 else "0"
            )
    return df


def _find_mods_and_sess(bids_dir: Path) -> dict:
    """Find all the modalities and sessions available for a given BIDS dataset.

    Parameters
    ----------
    bids_dir : Path
        The path to the BIDS dataset.

    Returns
    -------
    mods_dict : dict
        A dictionary that stores the sessions and modalities found and has the following structure.
        {
            'sessions': ['ses-M000', 'ses-M018'],
            'fmap': ['fmap'],
            'anat': ['flair', 't1w'],
            'func': ['func_task-rest'],
            'dwi': ['dwi']
        }
    """
    from collections import defaultdict

    mods_dict = defaultdict(set)

    for sub_path in bids_dir.glob("*sub-*"):
        for session in sub_path.glob("*ses-*"):
            mods_dict["sessions"].add(session.name)
            for modality in session.glob("*/"):
                if modality.name == "func":
                    for func_path in modality.glob("*bold.nii.gz"):
                        func_task = func_path.name.split("_")[2]
                        mods_dict["func"].add(f"func_{func_task}")
                if modality.name == "dwi":
                    mods_dict["dwi"].add("dwi")
                if modality.name == "fmap":
                    mods_dict["fmap"].add("fmap")
                if modality.name == "pet":
                    for pet_path in modality.glob("*pet.nii.gz"):
                        pet_name = pet_path.name.split(".")[0]
                        pet_acq = pet_name.split("_")[2]
                        mods_dict["pet"].add(f"pet_{pet_acq}")
                if modality.name == "anat":
                    for anat_file in modality.glob("*"):
                        if ".nii.gz" in anat_file.name:
                            anat_name = anat_file.name.replace(".nii.gz", "")
                            anat_ext = "nii.gz"
                        else:
                            anat_name = anat_file.stem
                            anat_ext = anat_file.suffix.lstrip(".")
                        if anat_ext != "json":
                            file_parts = anat_name.split("_")
                            anat_type = file_parts[-1].lower()
                            mods_dict["anat"].add(anat_type)

    return mods_dict



"""This module handles stream and log redirection."""
import warnings
from enum import IntEnum
from typing import Optional, Type, Union

__all__ = [
    "LoggingLevel",
    "get_logging_level",
    "cprint",
    "log_and_raise",
    "log_and_warn",
]


class LoggingLevel(IntEnum):
    DEBUG = 1
    INFO = 2
    WARNING = 3
    ERROR = 4
    CRITICAL = 5


def get_logging_level(level: Union[str, int, LoggingLevel]) -> LoggingLevel:
    if isinstance(level, str):
        if level == "debug":
            return LoggingLevel.DEBUG
        if level == "info":
            return LoggingLevel.INFO
        if level == "warning":
            return LoggingLevel.WARNING
        if level == "error":
            return LoggingLevel.ERROR
        if level == "critical":
            return LoggingLevel.CRITICAL
        raise ValueError(f"Logging level {level} is not valid.")
    return LoggingLevel(level)


def cprint(msg: str, lvl: Optional[Union[str, int, LoggingLevel]] = None) -> None:
    """Print message to the console at the desired logging level.

    Parameters
    ----------
    msg : str
        The message to log.

    lvl : str or int, or LoggingLevel, optional
        The logging level:
            - 1 = "debug"
            - 2 = "info"
            - 3 = "warning"
            - 4 = "error"
            - 5 = "critical"
        The default value is "info".
    """
    from logging import getLogger

    lvl = lvl or LoggingLevel.INFO
    lvl = get_logging_level(lvl)
    logger = getLogger("clinica")

    if lvl == LoggingLevel.DEBUG:
        logger.debug(msg=msg)
    elif lvl == LoggingLevel.INFO:
        logger.info(msg=msg)
    elif lvl == LoggingLevel.WARNING:
        logger.warning(msg=msg)
    elif lvl == LoggingLevel.ERROR:
        logger.error(msg=msg)
    elif lvl == LoggingLevel.CRITICAL:
        logger.critical(msg=msg)
    else:
        pass


def log_and_raise(message: str, error_type: Type[Exception]):
    """Log the error message using cprint and raise.

    Parameters
    ----------
    message : str
        The error message.

    error_type : Exception
        The error type to raise.
    """
    cprint(message, lvl=LoggingLevel.ERROR)
    raise error_type(message)


def log_and_warn(message: str, warning_type: Type[Warning]):
    """Log the warning message using cprint and give it to the user.

    Parameters
    ----------
    message : str
        The warning message.

    warning_type : Warning
        The warning type to use.
    """
    cprint(message, lvl=LoggingLevel.WARNING)
    warnings.warn(message, warning_type)



"""This module contains Nipype tasks used by the DWIDTI pipeline.

Nipype tasks must be 'self-contained' such that only primitive type hints can be
used. The tasks are simple wrappers around a properly implemented Python function.
"""


def compute_statistics_on_atlases_task(
    registered_map: str, name_map: str, dwi_preprocessed_file: str
) -> list:
    from pathlib import Path

    from clinica.pipelines.dwi.dti.utils import compute_statistics_on_atlases

    return compute_statistics_on_atlases(
        Path(registered_map),
        name_map,
        Path(dwi_preprocessed_file),
    )


def get_caps_filenames_task(caps_dwi_filename: str) -> tuple:
    from pathlib import Path

    from clinica.pipelines.dwi.dti.utils import get_caps_filenames

    names = get_caps_filenames(Path(caps_dwi_filename))

    return (
        names["bids_source"],
        names["diffmodel"],
        names["FA"],
        names["MD"],
        names["AD"],
        names["RD"],
        names["DECFA"],
    )


def rename_into_caps_task(
    in_caps_dwi: str,
    in_norm_fa: str,
    in_norm_md: str,
    in_norm_ad: str,
    in_norm_rd: str,
    in_b_spline_transform: str,
    in_affine_matrix: str,
) -> tuple:
    from pathlib import Path

    from clinica.pipelines.dwi.dti.utils import rename_into_caps

    return rename_into_caps(
        Path(in_caps_dwi),
        Path(in_norm_fa),
        Path(in_norm_md),
        Path(in_norm_ad),
        Path(in_norm_rd),
        Path(in_b_spline_transform),
        Path(in_affine_matrix),
    )



from enum import Enum
from typing import Optional

import pandas as pd

from ._base import GLM
from ._correlation import CorrelationGLM
from ._group import GroupGLM, GroupGLMWithInteraction

__all__ = ["GLMModelType", "create_glm_model"]


class GLMModelType(str, Enum):
    """Supported types of GLM models."""

    CORRELATION = "correlation"
    GROUP_COMPARISON = "group_comparison"


def create_glm_model(
    glm_type: GLMModelType,
    design: str,
    df: pd.DataFrame,
    contrast: str,
    feature_label: str,
    group_label: Optional[str] = "group",
    fwhm: Optional[int] = 20,
    threshold_uncorrected_pvalue: Optional[float] = 0.001,
    threshold_corrected_pvalue: Optional[float] = 0.05,
    cluster_threshold: Optional[float] = 0.001,
) -> GLM:
    """Factory method for building a GLM model instance corresponding to the
    provided type and design matrix.

    Parameters
    ----------
    glm_type : GLMModelType
        The type of GLM to be created.

    design : str
        The design matrix specified in string format.
        If this contains a "*", it will be interpreted as an interaction effect.

    df : pd.DataFrame
        The subjects DataFrame.

    contrast : str
        The contrast specified in string format.

    feature_label : str
        The label used for building output filenames.

    group_label : str, optional
        The label to use for group GLM models. Default="group".

    fwhm : int, optional
        The smoothing FWHM. This is used in the output file names.
        Default=20.

    threshold_uncorrected_pvalue : float, optional
        The threshold to be used with uncorrected P-values. Default=0.001.

    threshold_corrected_pvalue : float, optional
        The threshold to be used with corrected P-values. Default=0.05.

    cluster_threshold : float, optional
        The threshold to be used to declare clusters as significant. Default=0.001.

    Returns
    -------
    model : GLM
        An instance of the `GLM` class.

    Raises
    ------
    ValueError
        If the glm_type is not supported.
    """
    from clinica.utils.stream import cprint

    cprint(
        msg=f"The GLM model is: {design} and the GLM type is: {glm_type}",
        lvl="info",
    )
    params = {
        "group_label": group_label,
        "fwhm": fwhm,
        "threshold_uncorrected_pvalue": threshold_uncorrected_pvalue,
        "threshold_corrected_pvalue": threshold_corrected_pvalue,
        "cluster_threshold": cluster_threshold,
    }
    if glm_type == GLMModelType.CORRELATION:
        return CorrelationGLM(design, df, feature_label, contrast, **params)
    if glm_type == GLMModelType.GROUP_COMPARISON:
        if "*" in design:
            return GroupGLMWithInteraction(
                design, df, feature_label, contrast, **params
            )
        return GroupGLM(design, df, feature_label, contrast, **params)



import numpy as np
import pandas as pd

from ._utils import check_column_in_df, is_categorical

__all__ = [
    "Contrast",
    "CorrelationContrast",
    "GroupContrast",
    "GroupContrastWithInteraction",
]


class Contrast:
    def __init__(self, name: str, built_contrast: pd.Series):
        self.name = name
        self.built_contrast = built_contrast


class CorrelationContrast(Contrast):
    def __init__(
        self, name: str, absolute_name: str, built_contrast: pd.Series, sign: str
    ):
        super().__init__(name, built_contrast)
        self.sign = sign
        self.absolute_name = absolute_name

    @classmethod
    def from_string(cls, contrast: str, df: pd.DataFrame):
        absolute_contrast_name = contrast
        contrast_sign = "positive"
        if contrast.startswith("-"):
            absolute_contrast_name = contrast[1:].lstrip()
            contrast_sign = "negative"
        check_column_in_df(df, absolute_contrast_name)
        built_contrast = df[absolute_contrast_name]
        if contrast_sign == "negative":
            built_contrast *= -1
        return cls(contrast, absolute_contrast_name, built_contrast, contrast_sign)


class GroupContrastWithInteraction(Contrast):
    def __init__(self, name: str, built_contrast: pd.Series):
        super().__init__(name, built_contrast)

    @classmethod
    def from_string(cls, contrast: str, df: pd.DataFrame):
        contrast_elements = [_.strip() for _ in contrast.split("*")]
        for contrast_element in contrast_elements:
            check_column_in_df(df, contrast_element)
        categorical = [is_categorical(df, _) for _ in contrast_elements]
        if len(contrast_elements) != 2 or sum(categorical) != 1:
            raise ValueError(
                "The contrast must be an interaction between one continuous "
                "variable and one categorical variable. Your contrast contains "
                f"the following variables : {contrast_elements}"
            )
        idx = 0 if categorical[0] else 1
        categorical_contrast = contrast_elements[idx]
        continue_contrast = contrast_elements[(idx + 1) % 2]
        group_values = np.unique(df[categorical_contrast])
        built_contrast = df[continue_contrast].where(
            df[categorical_contrast] == group_values[0], 0
        ) - df[continue_contrast].where(df[categorical_contrast] == group_values[1], 0)
        return cls(contrast, built_contrast)


class GroupContrast(Contrast):
    def __int__(self, name: str, built_contrast: pd.Series):
        super().__init__(name, built_contrast)

    @classmethod
    def from_string(cls, contrast: str, df: pd.DataFrame, sign: str):
        check_column_in_df(df, contrast)
        if not is_categorical(df, contrast):
            raise ValueError(
                "Contrast should refer to a categorical variable for group comparison. "
                "Please select 'correlation' for 'glm_type' otherwise."
            )
        group_values = np.unique(df[contrast])
        i, j = (0, 1) if sign == "positive" else (1, 0)
        contrast_name = f"{group_values[j]}-lt-{group_values[i]}"
        built_contrast = (df[contrast] == group_values[i]).astype(int) - (
            df[contrast] == group_values[j]
        ).astype(int)

        return cls(contrast_name, built_contrast)



from pathlib import Path
from typing import Iterable, Optional, Union

from clinica.iotools.abstract_converter import Converter
from clinica.utils.filemanip import UserProvidedPath

from .adni_utils import ADNIModality

__all__ = ["convert"]


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    path_to_clinical: UserProvidedPath,
    clinical_data_only: bool = False,
    subjects: Optional[UserProvidedPath] = None,
    modalities: Optional[Iterable[Union[str, ADNIModality]]] = None,
    xml_path: Optional[UserProvidedPath] = None,
    force_new_extraction: bool = False,
    n_procs: Optional[int] = 1,
):
    from ..utils import validate_input_path

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    path_to_clinical = validate_input_path(path_to_clinical)
    if xml_path:
        xml_path = validate_input_path(xml_path)
    if subjects:
        subjects = validate_input_path(subjects)
    modalities = _validate_input_modalities(modalities)

    adni_to_bids = AdniToBids()
    adni_to_bids.check_adni_dependencies()

    if not clinical_data_only:
        adni_to_bids.convert_images(
            source_dir=path_to_dataset,
            clinical_dir=path_to_clinical,
            dest_dir=bids_dir,
            modalities=modalities,
            subjects=subjects,
            force_new_extraction=force_new_extraction,
            n_procs=n_procs,
        )
    adni_to_bids.convert_clinical_data(
        clinical_data_dir=path_to_clinical,
        out_path=bids_dir,
        clinical_data_only=clinical_data_only,
        subjects=subjects,
        xml_path=xml_path,
    )


class AdniToBids(Converter):
    @classmethod
    def check_adni_dependencies(cls) -> None:
        """Check the dependencies of ADNI converter."""
        from clinica.utils.check_dependency import ThirdPartySoftware, check_software

        check_software(ThirdPartySoftware.DCM2NIIX)

    def convert_clinical_data(
        self,
        clinical_data_dir: Path,
        out_path: Path,
        clinical_data_only: bool = False,
        subjects: Optional[Path] = None,
        xml_path: Optional[Path] = None,
    ):
        """Convert the clinical data of ADNI specified into the file clinical_specifications_adni.xlsx.

        Args:
            clinical_data_dir:  path to the clinical data directory
            out_path: path to the BIDS directory
            clinical_data_only: process clinical data only
            subjects: restrict processing to this manifest of subjects
            xml_path: path to the XML metadata files
        """
        from clinica.iotools.bids_utils import (
            StudyName,
            create_participants_df,
            get_bids_subjs_list,
            get_bids_subjs_paths,
            write_modality_agnostic_files,
        )
        from clinica.utils.stream import cprint

        from .adni_json import create_json_metadata
        from .adni_utils import (
            correct_diagnosis_sc_adni3,
            create_adni_scans_files,
            create_adni_sessions_dict,
        )

        clinical_specifications_folder = Path(__file__).parents[0] / "specifications"
        if not out_path.exists():
            msg = f"BIDS folder {out_path} not found."
            cprint(msg, lvl="error")
            raise FileNotFoundError(msg)
        conversion_path = out_path / "conversion_info"

        if clinical_data_only:
            bids_ids, bids_subjects_paths = _get_bids_subjects_info(
                clinical_data_dir=clinical_data_dir,
                out_path=out_path,
                subjects=subjects,
            )
        else:
            bids_ids = get_bids_subjs_list(out_path)
            bids_subjects_paths = get_bids_subjs_paths(out_path)

        # -- Creation of modality agnostic files --
        cprint("Creating modality agnostic files...", lvl="info")
        readme_data = {
            "link": "http://adni.loni.usc.edu",
            "desc": (
                "ADNI is a global research effort that actively supports the investigation and development of "
                "treatments that slow or stop the progression of Alzheimer's disease (AD).This multisite, longitudinal "
                "study assesses clinical, imaging, genetic and biospecimen biomarkers through the process of normal "
                "aging to mild cognitive impairment (MCI) and AD dementia.With established, standardized methods for "
                "imaging and biomarker collection and analysis, ADNI facilitates a way for scientists to conduct "
                "cohesive research and share compatible data with other researchers around the world."
            ),
        }
        write_modality_agnostic_files(
            study_name=StudyName.ADNI,
            readme_data=readme_data,
            bids_dir=out_path,
        )
        # -- Creation of participant.tsv --
        cprint("Creating participants.tsv...", lvl="info")
        participants_df = create_participants_df(
            StudyName.ADNI,
            clinical_specifications_folder,
            clinical_data_dir,
            bids_ids,
        )
        # Replace the original values with the standard defined by the AramisTeam
        participants_df["sex"] = participants_df["sex"].replace("Male", "M")
        participants_df["sex"] = participants_df["sex"].replace("Female", "F")

        # Correction of diagnosis_sc for ADNI3 participants
        participants_df = correct_diagnosis_sc_adni3(clinical_data_dir, participants_df)
        participants_df.to_csv(
            out_path / "participants.tsv",
            sep="\t",
            index=False,
            encoding="utf-8",
        )
        # -- Creation of sessions.tsv --
        cprint("Creating sessions files...", lvl="info")
        create_adni_sessions_dict(
            bids_ids,
            clinical_specifications_folder,
            clinical_data_dir,
            bids_subjects_paths,
        )
        # -- Creation of scans files --
        if conversion_path.exists():
            cprint("Creating scans files...", lvl="info")
            create_adni_scans_files(conversion_path, bids_subjects_paths)

        if xml_path is not None:
            if xml_path.exists():
                create_json_metadata(bids_subjects_paths, bids_ids, xml_path)
            else:
                cprint(
                    msg=(
                        f"Clinica was unable to find {xml_path}, "
                        "skipping xml metadata extraction."
                    ),
                    lvl="warning",
                )

    def convert_images(
        self,
        source_dir: Path,
        clinical_dir: Path,
        dest_dir: Path,
        modalities: Iterable[ADNIModality],
        subjects: Optional[Path] = None,
        force_new_extraction: bool = False,
        n_procs: Optional[int] = 1,
    ):
        """Convert the images of ADNI.

        Args:
            source_dir: path to the ADNI directory
            clinical_dir: path to the clinical data directory
            dest_dir: path to the BIDS directory
            subjects: Path to list of subjects to process
            modalities: modalities to convert (T1, PET_FDG, PET_AMYLOID, PET_TAU, DWI, FLAIR, fMRI)
            force_new_extraction: if given pre-existing images in the BIDS directory will be erased and extracted again.
        """
        from clinica.utils.stream import cprint

        from .adni_utils import get_subjects_list
        from .modality_converters import modality_converter_factory

        modalities = modalities or ADNIModality
        subjects = get_subjects_list(source_dir, clinical_dir, subjects)

        # Create the output folder if is not already existing
        (dest_dir / "conversion_info").mkdir(parents=True, exist_ok=True)
        version_number = len([f for f in (dest_dir / "conversion_info").iterdir()])
        conversion_dir = dest_dir / "conversion_info" / f"v{version_number}"
        conversion_dir.mkdir(parents=True, exist_ok=True)
        cprint(f"Destination folder = {dest_dir}", lvl="debug")

        for modality in modalities:
            for converter in modality_converter_factory(modality):
                converter(
                    source_dir=source_dir,
                    csv_dir=clinical_dir,
                    destination_dir=dest_dir,
                    conversion_dir=conversion_dir,
                    subjects=subjects,
                    mod_to_update=force_new_extraction,
                    n_procs=n_procs,
                )


def _validate_input_modalities(
    modalities: Optional[Iterable[str]] = None,
) -> tuple[ADNIModality, ...]:
    if modalities:
        return tuple((ADNIModality(modality) for modality in modalities))
    return tuple(ADNIModality)


def _get_bids_subjects_info(
    clinical_data_dir: Path,
    out_path: Path,
    subjects: Optional[Path] = None,
) -> tuple[list[str], list[Path]]:
    from clinica.iotools.bids_utils import StudyName, bids_id_factory

    from .adni_utils import load_clinical_csv

    # Read optional list of participants.
    if subjects:
        subjects = [subject for subject in subjects.read_text().split("\n") if subject]
    # Load all participants from ADNIMERGE.
    adni_merge = load_clinical_csv(clinical_data_dir, "ADNIMERGE")
    participants = adni_merge["PTID"].unique()
    # Filter participants if requested.
    participants = sorted(participants & subjects if subjects else participants)
    # Compute their corresponding BIDS IDs and paths.
    bids_ids = [
        bids_id_factory(StudyName.ADNI).from_original_study_id(p) for p in participants
    ]
    bids_paths = [out_path / bids_id for bids_id in bids_ids]

    return bids_ids, bids_paths



from pathlib import Path
from typing import Callable, Iterable, Optional, Union

from clinica.iotools.converters.adni_to_bids.adni_utils import (
    ADNIModality,
    ADNIModalityConverter,
)

__all__ = ["converter_factory", "modality_converter_factory"]


ConverterInterface = Callable[
    [Path, Path, Path, Path, Optional[Iterable[str]], bool, int], None
]


def _get_converters_for_modality(
    modality: ADNIModality,
) -> Iterable[ADNIModalityConverter]:
    """Map a user-facing modality from ADNI to an iterable of adni modality converters."""
    if modality == ADNIModality.T1:
        return [ADNIModalityConverter.T1]
    if modality == ADNIModality.PET_FDG:
        return [ADNIModalityConverter.PET_FDG, ADNIModalityConverter.PET_FDG_UNIFORM]
    if modality == ADNIModality.PET_AMYLOID:
        return [ADNIModalityConverter.PET_PIB, ADNIModalityConverter.PET_AV45]
    if modality == ADNIModality.PET_TAU:
        return [ADNIModalityConverter.PET_TAU]
    if modality == ADNIModality.DWI:
        return [ADNIModalityConverter.DWI]
    if modality == ADNIModality.FLAIR:
        return [ADNIModalityConverter.FLAIR]
    if modality == ADNIModality.FMRI:
        return [ADNIModalityConverter.FMRI]
    if modality == ADNIModality.FMAP:
        return [ADNIModalityConverter.FMAP]


def converter_factory(converter: ADNIModalityConverter) -> ConverterInterface:
    """Returns the converter associated with the provided ADNIModalityConverter variant."""
    if converter == ADNIModalityConverter.T1:
        from ._t1 import convert_t1

        return convert_t1
    if converter == ADNIModalityConverter.PET_FDG:
        from ._fdg_pet import convert_fdg_pet

        return convert_fdg_pet
    if converter == ADNIModalityConverter.PET_FDG_UNIFORM:
        from ._fdg_pet import convert_fdg_pet_uniform

        return convert_fdg_pet_uniform
    if converter == ADNIModalityConverter.PET_PIB:
        from ._pib_pet import convert_pib_pet

        return convert_pib_pet
    if converter == ADNIModalityConverter.PET_AV45:
        from ._av45_fbb_pet import convert_av45_fbb_pet

        return convert_av45_fbb_pet
    if converter == ADNIModalityConverter.PET_TAU:
        from ._tau_pet import convert_tau_pet

        return convert_tau_pet
    if converter == ADNIModalityConverter.DWI:
        from ._dwi import convert_dwi

        return convert_dwi
    if converter == ADNIModalityConverter.FLAIR:
        from ._flair import convert_flair

        return convert_flair
    if converter == ADNIModalityConverter.FMRI:
        from ._fmri import convert_fmri

        return convert_fmri
    if converter == ADNIModalityConverter.FMAP:
        from ._fmap import convert_fmap

        return convert_fmap


def modality_converter_factory(
    modality: Union[str, ADNIModality],
) -> Iterable[ConverterInterface]:
    """Returns an iterable of ADNI converters based on the provided modality."""
    return [
        converter_factory(converter)
        for converter in _get_converters_for_modality(ADNIModality(modality))
    ]



import click


class RegistrationOrderGroup(click.Group):
    """CLI group which lists commands by order or registration."""

    def list_commands(self, ctx):
        return self.commands.keys()


@click.group(cls=RegistrationOrderGroup, name="run")
def cli() -> None:
    """Run pipelines on BIDS and CAPS datasets."""
    pass


if __name__ == "__main__":
    cli()



"""Convert the NIFD dataset into BIDS."""

from typing import Optional

from clinica.utils.filemanip import UserProvidedPath

__all__ = ["convert"]


def convert(
    path_to_dataset: UserProvidedPath,
    bids_dir: UserProvidedPath,
    path_to_clinical: UserProvidedPath,
    subjects: Optional[UserProvidedPath] = None,
    n_procs: Optional[int] = 1,
    **kwargs,
):
    """Convert the entire dataset in BIDS.

    Scans available files in the path_to_dataset,
    identifies the patients that have images described by the JSON file,
    converts the image with the highest quality for each category.
    """
    from clinica.iotools.bids_utils import StudyName, write_modality_agnostic_files
    from clinica.iotools.converters.factory import get_converter_name
    from clinica.utils.check_dependency import ThirdPartySoftware, check_software
    from clinica.utils.stream import cprint

    from ..utils import validate_input_path
    from .nifd_utils import (
        dataset_to_bids,
        read_clinical_data,
        read_imaging_data,
        write_bids,
    )

    path_to_dataset = validate_input_path(path_to_dataset)
    bids_dir = validate_input_path(bids_dir, check_exist=False)
    path_to_clinical = validate_input_path(path_to_clinical)
    check_software(ThirdPartySoftware.DCM2NIIX)
    if subjects:
        cprint(
            (
                f"Subject filtering is not yet implemented in {get_converter_name(StudyName.NIFD)} converter. "
                "All subjects available will be converted."
            ),
            lvl="warning",
        )
    if n_procs != 1:
        cprint(
            f"{get_converter_name(StudyName.NIFD)} converter does not support multiprocessing yet. n_procs set to 1.",
            lvl="warning",
        )
    clinical_data = read_clinical_data(path_to_clinical)
    imaging_data = read_imaging_data(path_to_dataset)
    participants, sessions, scans = dataset_to_bids(
        imaging_data=imaging_data,
        clinical_data=clinical_data,
    )
    write_bids(
        to=bids_dir,
        participants=participants,
        sessions=sessions,
        scans=scans,
    )
    readme_data = {
        "link": "https://ida.loni.usc.edu/home/projectPage.jsp?project=NIFD&page=HOME&subPage=OVERVIEW_PR#",
        "desc": (
            "NIFD is the nickname for the frontotemporal lobar degeneration neuroimaging initiative "
            "(FTLDNI, AG032306), which was funded by the NIA and NINDS to characterize longitudinal clinical and "
            "imaging changes in FTLD.The imaging and clinical methods are the same for NIFD and for the 4-Repeat "
            "Tauopathy Neuroimaging Initiative (4RTNI), which is also available for download from LONI. Controls for "
            "NIFD are the same controls as those collected for 4RTNI."
        ),
    }
    write_modality_agnostic_files(
        study_name=StudyName.NIFD,
        readme_data=readme_data,
        bids_dir=bids_dir,
    )
    cprint("Conversion to BIDS succeeded.", lvl="info")



from typing import Any

import pydra
from nipype.algorithms.misc import Gunzip
from nipype.interfaces.petpvc import PETPVC
from nipype.interfaces.spm import Coregister, DARTELNorm2MNI, Smooth
from nipype.interfaces.spm.utils import Reslice
from pydra.engine import Workflow
from pydra.tasks.nipype1.utils import Nipype1Task

from clinica.pydra.engine import clinica_io
from clinica.pydra.pet_volume.tasks import (
    apply_binary_mask_task,
    atlas_statistics_task,
    create_binary_mask_task,
    create_pvc_mask_task,
    get_psf_task,
    normalize_to_reference_task,
    pet_pvc_name_task,
)


def _check_pipeline_parameters(parameters: dict) -> dict:
    """Check the parameters passed to the pipeline.

    Parameters
    ----------
    parameters : dict
        Dictionary of parameters to analyze.

    Returns
    -------
    dict :
        Cleaned dictionary of parameters.
    """
    from clinica.utils.atlas import T1AndPetVolumeAtlasName
    from clinica.utils.group import check_group_label

    parameters.setdefault("group_label", None)
    check_group_label(parameters["group_label"])
    if "acq_label" not in parameters.keys():
        raise KeyError("Missing compulsory acq_label key in pipeline parameter.")
    parameters.setdefault("pvc_psf_tsv", None)
    parameters["apply_pvc"] = parameters["pvc_psf_tsv"] is not None
    parameters.setdefault("mask_tissues", [1, 2, 3])
    parameters.setdefault("mask_threshold", 0.3)
    parameters.setdefault("pvc_mask_tissues", [1, 2, 3])
    parameters.setdefault("smooth", 8.0)
    parameters.setdefault("atlases", T1AndPetVolumeAtlasName)
    return parameters


@clinica_io
def build_core_workflow(name: str = "core", parameters: dict = {}) -> Workflow:
    """Build the core workflow for the PET Volume pipeline.

    Parameters
    ----------
    name : str, optional
        The name of the workflow. Default="core".

    parameters : dict, optional
        Optional dictionary of parameters to be used
        within the workflow.
        Default={}.

    Returns
    -------
    wf : Workflow
        The core workflow.
    """
    from pydra.tasks import petpvc

    from clinica.pydra.shared_workflows.smoothing import build_smoothing_workflow
    from clinica.pydra.utils import sanitize_fwhm
    from clinica.utils.pet import get_suvr_mask
    from clinica.utils.spm import use_spm_standalone_if_available

    use_spm_standalone_if_available()

    parameters = _check_pipeline_parameters(parameters)

    input_spec = pydra.specs.SpecInfo(
        name="Input",
        fields=[
            ("_graph_checksums", Any),
            ("T1w", dict, {}, {"mandatory": True}),
            ("pet", dict, {"tracer": parameters["acq_label"]}, {"mandatory": True}),
            (
                "mask_tissues",
                dict,
                {"tissue_number": parameters["mask_tissues"], "modulation": False},
                {"mandatory": True},
            ),
            (
                "flow_fields",
                dict,
                {"group_label": parameters["group_label"]},
                {"mandatory": True},
            ),
            (
                "dartel_template",
                dict,
                {"group_label": parameters["group_label"]},
                {"mandatory": True},
            ),
            (
                "pvc_mask_tissues",
                dict,
                {"tissue_number": parameters["pvc_mask_tissues"]},
                {"mandatory": False},
            ),
        ],
        bases=(pydra.specs.BaseSpec,),
    )

    wf = Workflow(
        name,
        input_spec=input_spec,
    )
    wf.split(("pet", "T1w", "flow_fields", "pvc_mask_tissues", "mask_tissues"))
    compressed_inputs = [
        "pet",
        "T1w",
        "flow_fields",
        "dartel_template",
    ]
    # Unzipping
    for input_name in compressed_inputs:
        wf.add(
            Nipype1Task(
                name=f"unzip_{input_name}",
                interface=Gunzip(),
                in_file=getattr(wf.lzin, input_name),
            )
        )
    wf.add(
        Nipype1Task(
            name="unzip_mask_tissues",
            interface=Gunzip(),
            in_file=wf.lzin.mask_tissues,
        )
        .split("in_file")
        .combine("in_file")
    )
    wf.add(
        Nipype1Task(
            name="unzip_reference_mask",
            interface=Gunzip(),
            in_file=get_suvr_mask(parameters["suvr_reference_region"]),
        )
    )

    # Coregister PET into T1 native space
    coreg_pet_t1 = Nipype1Task(
        name="coreg_pet_t1",
        interface=Coregister(),
    )
    coreg_pet_t1.inputs.source = wf.unzip_pet.lzout.out_file
    coreg_pet_t1.inputs.target = wf.unzip_T1w.lzout.out_file
    wf.add(coreg_pet_t1)

    # Spatially normalize PET into MNI
    dartel_mni_reg = Nipype1Task(
        name="dartel_mni_reg",
        interface=DARTELNorm2MNI(),
    )
    dartel_mni_reg.inputs.modulate = False
    dartel_mni_reg.inputs.fwhm = 0
    dartel_mni_reg.inputs.flowfield_files = wf.unzip_flow_fields.lzout.out_file
    dartel_mni_reg.inputs.template_file = wf.unzip_dartel_template.lzout.out_file
    dartel_mni_reg.inputs.apply_to_files = wf.coreg_pet_t1.lzout.coregistered_source
    wf.add(dartel_mni_reg)

    # Reslice reference region mask into PET
    reslice = Nipype1Task(
        name="reslice",
        interface=Reslice(),
    )
    reslice.inputs.in_file = wf.unzip_reference_mask.lzout.out_file
    reslice.inputs.space_defining = wf.dartel_mni_reg.lzout.normalized_files
    wf.add(reslice)

    # Normalize PET values according to reference region
    wf.add(
        normalize_to_reference_task(
            name="norm_to_ref",
            interface=normalize_to_reference_task,
            pet_image=wf.dartel_mni_reg.lzout.normalized_files,
            region_mask=wf.reslice.lzout.out_file,
        )
    )

    # Create binary mask from segmented tissues
    wf.add(
        create_binary_mask_task(
            name="binary_mask",
            interface=create_binary_mask_task,
            tissues=wf.unzip_mask_tissues.lzout.out_file,
            threshold=parameters["mask_threshold"],
        )
    )

    # Mask PET image
    wf.add(
        apply_binary_mask_task(
            name="apply_mask",
            interface=apply_binary_mask_task,
            image=wf.norm_to_ref.lzout.suvr_pet_path,
            binary_mask=wf.binary_mask.lzout.out_mask,
        )
    )

    # Smoothing
    if parameters["smooth"] is not None:
        smoothing_wf = build_smoothing_workflow(
            "smoothing_workflow",
            sanitize_fwhm(parameters["smooth"]),
        )
        smoothing_wf.inputs.input_file = wf.apply_mask.lzout.masked_image_path
        wf.add(smoothing_wf)

    # Atlas Statistics
    wf.add(
        atlas_statistics_task(
            name="atlas_stats_node",
            interface=atlas_statistics_task,
            in_image=wf.norm_to_ref.lzout.suvr_pet_path,
            in_atlas_list=parameters["atlases"],
        )
    )

    # PVC
    if parameters["apply_pvc"]:
        # Unzipping
        wf.add(
            Nipype1Task(
                name="unzip_pvc_mask_tissues",
                interface=Gunzip(),
                in_file=wf.lzin.pvc_mask_tissues,
            )
            .split("in_file")
            .combine("in_file")
        )

        # Creating Mask to use in PVC
        wf.add(
            create_pvc_mask_task(
                name="pvc_mask",
                interface=create_pvc_mask_task,
                tissues=wf.unzip_pvc_mask_tissues.lzout.out_file,
            )
        )

        # Build PET PVC name for PET PVC task
        wf.add(
            pet_pvc_name_task(
                name="pet_pvc_name",
                interface=pet_pvc_name_task,
                pet_image=wf.coreg_pet_t1.lzout.coregistered_source,
                pvc_method="RBV",
            )
        )

        # Retrieve PSF for PET PVC task
        wf.add(
            get_psf_task(
                name="get_psf_task",
                interface=get_psf_task,
                pvc_psf_tsv=parameters["pvc_psf_tsv"],
                pet_filename=wf.lzin.pet,
                pet_tracer=parameters["acq_label"],
            )
        )

        # PET PVC
        wf.add(
            petpvc.PETPVC(
                name="pvc",
                pvc_method="RBV",
                output_image=wf.pet_pvc_name.lzout.pet_pvc_path,
                input_image=wf.coreg_pet_t1.lzout.coregistered_source,
                mask_image=wf.pvc_mask.lzout.out_mask,
                fwhm_x=wf.get_psf_task.lzout.psf_x,
                fwhm_y=wf.get_psf_task.lzout.psf_y,
                fwhm_z=wf.get_psf_task.lzout.psf_z,
            )
        )

        # Spatially normalize PET into MNI
        dartel_mni_reg_pvc = Nipype1Task(
            name="dartel_mni_reg_pvc",
            interface=DARTELNorm2MNI(),
        )
        dartel_mni_reg_pvc.inputs.modulate = False
        dartel_mni_reg_pvc.inputs.fwhm = 0
        dartel_mni_reg_pvc.inputs.flowfield_files = wf.unzip_flow_fields.lzout.out_file
        dartel_mni_reg_pvc.inputs.template_file = (
            wf.unzip_dartel_template.lzout.out_file
        )
        dartel_mni_reg_pvc.inputs.apply_to_files = wf.pvc.lzout.output_image
        wf.add(dartel_mni_reg_pvc)

        # Reslice reference region mask into PET
        reslice_pvc = Nipype1Task(
            name="reslice_pvc",
            interface=Reslice(),
        )
        reslice_pvc.inputs.space_defining = wf.dartel_mni_reg_pvc.lzout.normalized_files
        reslice_pvc.inputs.in_file = wf.unzip_reference_mask.lzout.out_file
        wf.add(reslice_pvc)

        # Normalize PET values according to reference region
        wf.add(
            normalize_to_reference_task(
                name="norm_to_ref_pvc",
                interface=normalize_to_reference_task,
                pet_image=wf.dartel_mni_reg_pvc.lzout.normalized_files,
                region_mask=wf.reslice_pvc.lzout.out_file,
            )
        )

        # Mask PET image
        wf.add(
            apply_binary_mask_task(
                name="apply_mask_pvc",
                interface=apply_binary_mask_task,
                image=wf.norm_to_ref_pvc.lzout.suvr_pet_path,
                binary_mask=wf.binary_mask.lzout.out_mask,
            )
        )

        # Smoothing
        if parameters["smooth"] is not None:
            pvc_smoothing_wf = build_smoothing_workflow(
                "pvc_smoothing_workflow",
                sanitize_fwhm(parameters["smooth"]),
            )
            pvc_smoothing_wf.inputs.input_file = (
                wf.apply_mask_pvc.lzout.masked_image_path
            )
            wf.add(pvc_smoothing_wf)

        # Atlas Statistics
        wf.add(
            atlas_statistics_task(
                name="atlas_stats_pvc",
                interface=atlas_statistics_task,
                in_image=wf.norm_to_ref_pvc.lzout.suvr_pet_path,
                in_atlas_list=parameters["atlases"],
            )
        )

    output_connections = [
        ("pet_suvr_masked_smoothed", wf.smoothing_workflow.lzout.smoothed_files),
        ("pet_t1_native", wf.coreg_pet_t1.lzout.coregistered_source),
        ("pet_mni", wf.dartel_mni_reg.lzout.normalized_files),
        ("pet_suvr", wf.norm_to_ref.lzout.suvr_pet_path),
        ("binary_mask", wf.binary_mask.lzout.out_mask),
        ("pet_suvr_masked", wf.apply_mask.lzout.masked_image_path),
        ("atlas_statistics", wf.atlas_stats_node.lzout.atlas_statistics),
    ]

    if parameters["apply_pvc"]:
        output_connections += [
            (
                "pet_pvc_suvr_masked_smoothed",
                wf.pvc_smoothing_workflow.lzout.smoothed_files,
            ),
            ("pet_pvc", wf.pvc.lzout.output_image),
            ("pet_pvc_mni", wf.dartel_mni_reg_pvc.lzout.normalized_files),
            ("pet_pvc_suvr", wf.norm_to_ref_pvc.lzout.suvr_pet_path),
            ("pet_pvc_suvr_masked", wf.apply_mask_pvc.lzout.masked_image_path),
            ("pvc_atlas_statistics", wf.atlas_stats_pvc.lzout.atlas_statistics),
        ]

    wf.set_output(output_connections)

    return wf



from pathlib import Path

import pandas as pd

from clinica.utils.stream import cprint

__all__ = ["find_image_path"]


def find_image_path(
    images: pd.DataFrame,
    source_dir: Path,
    modality: str,
) -> pd.DataFrame:
    """
    For each image, the path to an existing image file or folder is created from image metadata.

    This function adds two columns to the input dataframe: 'Is_Dicom', and 'Path'.

    Args:
        images: List of images metadata
        source_dir: path to the ADNI directory
        modality: Imaging modality

    Returns: Dataframe containing metadata and existing paths
    """
    is_dicom = []
    image_folders = []
    for _, image in images.iterrows():
        image_path, dicom = _find_path_single_image(image, source_dir)
        is_dicom.append(dicom)
        image_folders.append(image_path)
        if image_path == "":
            cprint(
                msg=(
                    f"No {modality} image path found for subject {image.Subject_ID} in visit {image.VISCODE} "
                    f"with image ID {image.Image_ID}"
                ),
                lvl="info",
            )
    images.loc[:, "Is_Dicom"] = pd.Series(is_dicom, index=images.index)
    images.loc[:, "Path"] = pd.Series(image_folders, index=images.index)

    return images


def _find_path_single_image(image: pd.Series, source_dir: Path) -> tuple[str, bool]:
    path_to_sequence = source_dir / str(image["Subject_ID"])
    image_folder_path = ""
    is_dicom = True
    if (
        len(
            (
                found_images := [
                    f
                    for f in path_to_sequence.rglob(f"*_I{image['Image_ID']}.*")
                    if f.is_file()
                ]
            )
        )
        > 0
    ):
        if "dcm" in found_images[0].suffix:
            image_folder_path = str(found_images[0].parent)
        else:
            image_folder_path = str(found_images[0])
            is_dicom = False
    return image_folder_path, is_dicom



from pathlib import Path
from typing import Callable

from nilearn.surface import Mesh

from ._statistics import StatisticsResults

__all__ = ["StatisticsResultsPlotter"]


class StatisticsResultsPlotter:
    """Class responsible to plotting results of GLM fit.

    Attributes
    ----------
    output_file : PathLike
        Path to the output file.

    mesh : nilearn.surface.Mesh
        The mesh to be used for plotting results.
    """

    def __init__(self, output_file: Path, mesh: Mesh):
        self.output_file = output_file
        self.mesh = mesh
        self.plotting_extension = ".png"
        self.no_plot = {"coefficients"}  # Elements which should not be plotted

    def plot(self, result: StatisticsResults, method: str) -> None:
        """Plot the results.

        Parameters
        ----------
        result : StatisticsResults
            The results to be plotted.

        method : str
            The plotting method to use.
        """
        plotter = self._get_plotter(method)
        plotter(result)

    def _get_plotter(self, method: str) -> Callable[[StatisticsResults], None]:
        """Returns the plotting method from its name.

        Parameters
        ----------
        method : str
            Name of the plotting method to use.

        Returns
        -------
        Callable :
            Plotting method.
        """
        if method == "nilearn_plot_surf_stat_map":
            return self._plot_stat_maps
        raise NotImplementedError(f"Plotting method {method} is not implemented.")

    def _plot_stat_maps(self, result: StatisticsResults) -> None:
        """Wrapper around the `nilearn.plotting.plot_surf_stat_map` method.

        Parameters
        ----------
        result : StatisticsResults
            The results to plot.
        """
        from nilearn.plotting import plot_surf_stat_map

        from clinica.utils.stream import cprint

        for name, res in result.to_dict().items():
            if name not in self.no_plot:
                texture = res
                threshold = None
                plot_filename = (
                    str(self.output_file) + "_" + name + self.plotting_extension
                )
                if isinstance(res, dict):
                    texture = res["P"]
                    threshold = res["thresh"]
                cprint(msg=f"Saving plot to {plot_filename}", lvl="info")
                plot_surf_stat_map(
                    self.mesh,
                    texture,
                    threshold=threshold,
                    output_file=plot_filename,
                    title=name,
                )



__all__ = ["compute_atlas_task", "write_tsv_files_task"]


def compute_atlas_task(
    caps_directory: str, to_process_with_atlases: tuple, path_to_atlas: str
) -> tuple:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.atlas.utils import compute_atlas

    subject_dir, image_id, atlas = compute_atlas(
        Path(caps_directory),
        to_process_with_atlases,
        Path(path_to_atlas),
    )
    return str(subject_dir), image_id, atlas


def write_tsv_files_task(subject_dir: str, image_id: str, atlas: str) -> str:
    """Adapter for Nipype."""
    from pathlib import Path

    from clinica.pipelines.anatomical.freesurfer.atlas.utils import write_tsv_files

    return write_tsv_files(Path(subject_dir), image_id, atlas)



from pathlib import Path

__all__ = [
    "extract_participant_long_ids_from_filename",
    "init_input_node",
    "run_recon_all_base",
    "save_to_caps",
]


def extract_participant_long_ids_from_filename(
    caps_files: list[str],
) -> tuple[list[str], list[str]]:
    """TODO: Find a way to merge with utils/filemanip.py::extract_subjects_sessions_from_filename into one util."""
    import re

    caps_files = [
        re.search(r"(sub-[a-zA-Z0-9]+)_(long-[a-zA-Z0-9]+)", file).group()
        for file in caps_files
    ]
    split = [image_id.split("_") for image_id in caps_files]
    part_ids = [p_id[0] for p_id in split]
    long_ids = [l_id[1] for l_id in split]
    return part_ids, long_ids


def init_input_node(
    caps_dir: Path,
    participant_id: str,
    list_session_ids: list[str],
    output_dir: Path,
) -> tuple[str, Path, str]:
    """Initialize the pipeline.

    This function will create folders and symbolic links in SUBJECTS_DIR env variable for upcoming run of recon-all.

    Note (@alexis-g-icm):
        There currently (as of 22 Feb 2019) is a bug in FreeSurfer recon-all -base, which in some cases (e.g., only one
        time point), will crash as it's trying to write lines too long for the shell to handle. This is caused by
        the path to FreeSurfer SUBJECT_DIR being too long itself.

    The current function works around this issue by checking if there only is one session associated to a subject, and
    in that case, putting the SUBJECT_DIR inside the system temporary folder so that its path is as short as possible.
    """
    from tempfile import mkdtemp

    from clinica.compat import errno
    from clinica.utils.longitudinal import get_long_id
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_begin_image

    long_id = get_long_id(list_session_ids)
    image_id = f"{participant_id}_{long_id}"
    if len(list_session_ids) == 1:
        # Special case: When only one time point is used, 'recon-all -base' can failed
        # if the $SUBJECTS_DIR is too long ('Word too long.' error).
        # To circumvent this issue, we create a sym link in $(TMP) so that $SUBJECTS_DIR is a short path
        subjects_dir = Path(mkdtemp())
        cprint(
            msg=(
                f"{image_id.replace('_', ' | ')} has only one time point. "
                f"Needs to create a $SUBJECTS_DIR folder in {subjects_dir}"
            ),
            lvl="warning",
        )
    else:
        subjects_dir = output_dir / image_id
    subjects_dir.mkdir(parents=True, exist_ok=True)
    # Create symbolic links containing cross-sectional segmentation(s) in SUBJECTS_DIR so that recon-all can run
    for session_id in list_session_ids:
        cross_sectional_path = (
            caps_dir
            / "subjects"
            / participant_id
            / session_id
            / "t1"
            / "freesurfer_cross_sectional"
            / f"{participant_id}_{session_id}"
        )
        try:
            (subjects_dir / f"{participant_id}_{session_id}").symlink_to(
                cross_sectional_path
            )
        except FileExistsError as e:
            if e.errno != errno.EEXIST:  # EEXIST: folder already exists
                raise e
    flags = ""
    for session_id in list_session_ids:
        flags += f" -tp {participant_id}_{session_id}"
    print_begin_image(image_id)

    return image_id, subjects_dir, flags


def run_recon_all_base(subjects_dir: str, subject_id: str, flags, directive):
    """Run recon-all to create an unbiased template from all time points.

    Note:
    Unbiased template creation with FreeSurfer expects arguments to follow this syntax:
    recon-all -base <template_id> -tp <tp1_id> -tp <tp2_id> ... -all

    Currently, Nipype does not provide interface for "recon-all -base" case. As a result, ReconAll interface should be
    modified to handle this situation. In the meantime, the arguments of this function follows ReconAll.inputs names
    namely:
        - "-base <template_id>" is likely to be fed to ReconAll.inputs.subject_id or new input (e.g. template_id?)
        - "-tp <tp1_id> -tp <tp2_id> ..." is likely to  be fed to ReconAll.inputs.args
        - "-all" will be fed to ReconAll.inputs.directive

    See official documentation (https://surfer.nmr.mgh.harvard.edu/fswiki/LongitudinalProcessing) for details.
    """
    import subprocess

    recon_all_base_command = "recon-all -base {0} {1} -sd {2} {3}".format(
        subject_id, flags, subjects_dir, directive
    )
    subprocess_run_recon_all_base = subprocess.run(
        recon_all_base_command,
        shell=True,
        stdout=subprocess.DEVNULL,
        stderr=subprocess.DEVNULL,
    )
    if subprocess_run_recon_all_base.returncode != 0:
        raise ValueError("recon-all -base failed, returned non-zero code")

    return subject_id


def save_to_caps(
    source_dir: Path,
    image_id: str,
    list_session_ids: list[str],
    caps_dir: Path,
    overwrite_caps: bool = False,
) -> str:
    """Save `source_dir`/`image_id`/ to CAPS folder.

    This function copies FreeSurfer segmentation and regional_measures folder of `source_dir`/`image_id`/ to
    `caps_dir`/subjects/<participant_id>/<long_id>/freesurfer_unbiased_template/
    where `image_id` = <participant_id>_<long_id>.

    The `source_dir`/`image_id`/ folder should contain the following elements:
        - fsaverage, lh.EC_average and rh.EC_average symbolic links automatically generated by recon-all
        - symbolic links to cross-sectional segmentation(s) and unbiased template needed for recon-all
        - <participant_id>_<long_id>/ folder containing the FreeSurfer segmentation

    Notes:
        We do not need to check the line "finished without error" in scripts/recon-all.log.
        If an error occurs, it will be detected by Nipype and the next nodes (i.e. save_to_caps will not be called).
    """
    import shutil

    from clinica.utils.longitudinal import save_long_id
    from clinica.utils.stream import cprint
    from clinica.utils.ux import print_end_image

    participant_id = image_id.split("_")[0]
    long_id = image_id.split("_")[1]
    sessions_tsv_path = caps_dir.expanduser() / "subjects" / participant_id / long_id
    destination_dir = sessions_tsv_path / "freesurfer_unbiased_template"
    if not (sessions_tsv_path / f"{long_id}_sessions.tsv").is_file():
        save_long_id(list_session_ids, sessions_tsv_path, f"{long_id}_sessions.tsv")
    source_file = (
        source_dir.expanduser() / image_id / image_id / "mri" / "aparc+aseg.mgz"
    )
    destination_file = destination_dir / image_id / "mri" / "aparc+aseg.mgz"
    if source_file.is_file():
        if destination_file.is_file():
            if overwrite_caps:
                shutil.rmtree(destination_dir)
                shutil.copytree(
                    src=source_dir / image_id / image_id,
                    dst=destination_dir / image_id,
                    symlinks=True,
                )
        else:
            shutil.copytree(
                src=source_dir / image_id / image_id,
                dst=destination_dir / image_id,
                symlinks=True,
            )
        print_end_image(image_id)
    else:
        cprint(
            msg=(
                f"{image_id.replace('_', ' | ')}  does not contain "
                "mri/aseg+aparc.mgz file. Copy will be skipped."
            ),
            lvl="warning",
        )
    return image_id



import os
from pathlib import Path
from typing import Iterable

import numpy as np
import pandas as pd

from clinica.iotools.bids_utils import StudyName, get_bids_sess_list
from clinica.utils.stream import cprint

__all__ = ["create_sessions_dict", "write_sessions_tsv"]


def create_sessions_dict(
    clinical_data_dir: Path,
    bids_dir: Path,
    clinical_specifications_folder: Path,
    bids_ids: Iterable[str],
) -> dict:
    """Extract the information regarding the sessions and store them in a dictionary (session M000 only).

    Parameters
    ----------
    clinical_data_dir : Path
        The path to the input folder.

    bids_dir : Path
        The path to the BIDS directory.

    clinical_specifications_folder : Path
        The path to the clinical file.

    bids_ids : list of str
        The list of bids ids.

    Returns
    -------
    dict :
        Session dict.
    """

    location = f"{StudyName.OASIS.value} location"
    sessions = pd.read_csv(clinical_specifications_folder / "sessions.tsv", sep="\t")
    sessions_fields = sessions[StudyName.OASIS.value]
    field_location = sessions[location]
    sessions_fields_bids = sessions["BIDS CLINICA"]
    fields_dataset = []
    fields_bids = []
    sessions_dict = {}

    for i in range(0, len(sessions_fields)):
        if not pd.isnull(sessions_fields[i]):
            fields_bids.append(sessions_fields_bids[i])
            fields_dataset.append(sessions_fields[i])

    for i in range(0, len(sessions_fields)):
        # If the i-th field is available
        if not pd.isnull(sessions_fields[i]):
            # Load the file
            tmp = field_location[i].split("/")
            location = tmp[0]
            sheet = tmp[1] if len(tmp) > 1 else 0
            file_to_read_path = clinical_data_dir / location
            file_ext = os.path.splitext(location)[1]
            if file_ext == ".xlsx":
                file_to_read = pd.read_excel(file_to_read_path, sheet_name=sheet)
            elif file_ext == ".csv":
                file_to_read = pd.read_csv(file_to_read_path)
            else:
                raise ValueError(
                    f"Unknown file extension {file_ext}. Expecting either .xlsx or .csv."
                )

            for r in range(0, len(file_to_read.values)):
                # Extracts the subject ids columns from the dataframe
                subj_id = file_to_read.iloc[r]["ID"]
                if hasattr(subj_id, "dtype"):
                    if subj_id.dtype == np.int64:
                        subj_id = str(subj_id)
                # Removes all the - from
                subj_id_alpha = str(subj_id[0:3] + "IS" + subj_id[3] + subj_id[5:9])

                # Extract the corresponding BIDS id and create the output file if doesn't exist
                subj_bids = [s for s in bids_ids if subj_id_alpha in s]
                if subj_bids:
                    subj_bids = subj_bids[0]
                    subj_dir = bids_dir / subj_bids
                    session_names = get_bids_sess_list(subj_dir)
                    for s in session_names:
                        s_name = s.replace("ses-", "")
                        row = file_to_read.iloc[r]
                        if subj_bids not in sessions_dict:
                            sessions_dict.update({subj_bids: {}})
                        if s_name not in sessions_dict[subj_bids].keys():
                            sessions_dict[subj_bids].update({s_name: {"session_id": s}})
                        (sessions_dict[subj_bids][s_name]).update(
                            {sessions_fields_bids[i]: row[sessions_fields[i]]}
                        )

    return sessions_dict


def write_sessions_tsv(bids_dir: Path, sessions_dict: dict) -> None:
    """Create <participant_id>_sessions.tsv files.

    Basically writes the content of the function
    `clinica.iotools.bids_utils.create_sessions_dict` in several TSV files
    following the BIDS specification.

    Parameters
    ----------
    bids_dir : Path
        The path to the BIDS directory.

    sessions_dict : dict
        Dictionary containing sessions metadata.

        .. note::
            This is the output of the function
            `clinica.iotools.bids_utils.create_sessions_dict`.

    See also
    --------
    create_sessions_dict
    write_scans_tsv
    """
    for subject_path in bids_dir.glob("sub-*"):
        if subject_path.name in sessions_dict:
            session_df = pd.DataFrame.from_dict(
                sessions_dict[subject_path.name], orient="index"
            )
            cols = session_df.columns.tolist()
            cols = cols[-1:] + cols[:-1]
            session_df = session_df[cols]
        else:
            print(f"No session data available for {subject_path}")
            session_df = pd.DataFrame(columns=["session_id"])
            session_df["session_id"] = pd.Series("M000")
        session_df = session_df.set_index("session_id").fillna("n/a")
        session_df.to_csv(
            subject_path / f"{subject_path.name}_sessions.tsv",
            sep="\t",
            encoding="utf8",
        )



"""Common CLI arguments used by Clinica pipelines."""
import click

from clinica.utils.pet import SUVRReferenceRegion, Tracer

acq_label = click.argument(
    "acq_label",
    type=click.Choice(Tracer),
)

bids_directory = click.argument(
    "bids_directory",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
)

caps_directory = click.argument(
    "caps_directory",
    type=click.Path(writable=True, file_okay=False, resolve_path=True),
)

contrast = click.argument("contrast")

group_label = click.argument("group_label")

orig_input_data = click.argument(
    "orig_input_data",
    type=click.Choice(["t1-freesurfer", "pet-surface", "custom-pipeline"]),
)

orig_input_data_volume = click.argument(
    "orig_input_data_volume",
    type=click.Choice(["t1-volume", "pet-volume", "custom-pipeline"]),
)

orig_input_data_ml = click.argument(
    "orig_input_data_ml",
    type=click.Choice(["t1-volume", "pet-surface"]),
)

pvc_psf_tsv = click.argument(
    "pvc_psf_tsv",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)

subject_visits_with_covariates_tsv = click.argument(
    "subject_visits_with_covariates_tsv",
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
)

suvr_reference_region = click.argument(
    "suvr_reference_region",
    type=click.Choice(SUVRReferenceRegion),
)



from pathlib import Path

import click


@click.command(name="habs-to-bids")
@click.argument(
    "sourcedata",
    type=click.Path(exists=True, resolve_path=True, path_type=Path),
)
@click.argument(
    "rawdata",
    type=click.Path(writable=True, resolve_path=True, path_type=Path),
)
def cli(sourcedata: str, rawdata: str) -> None:
    """HABS to BIDS converter."""
    from .habs_to_bids import convert

    convert(sourcedata, rawdata)


if __name__ == "__main__":
    cli()



import abc
from pathlib import Path
from typing import Dict, List, Optional, Union

import numpy as np
import pandas as pd
from brainstat.stats.SLM import SLM
from brainstat.stats.terms import FixedEffect
from nilearn.surface import Mesh

from clinica.utils.stream import cprint

from ._contrast import Contrast
from ._utils import build_model, print_clusters
from .results import (
    StatisticsResults,
    StatisticsResultsPlotter,
    StatisticsResultsSerializer,
)

__all__ = ["GLM"]


class GLM:
    """This class implements the functionalities common to all GLM models
    used in the Clinica SurfaceStatistics pipeline.

    Attributes
    ----------
    design : str
        The design matrix specified in string format.
        If this contains a "*", it will be interpreted as an interaction effect.

    df : pd.DataFrame
        The subjects DataFrame.

    feature_label : str
        The label used for building output filenames.

    contrast : str
        The contrast specified in string format.

    fwhm : int, optional
        The smoothing FWHM. This is used in the output file names.
        Default=20.

    threshold_uncorrected_pvalue : float, optional
        The threshold to be used with uncorrected P-values. Default=0.001.

    threshold_corrected_pvalue : float, optional
        The threshold to be used with corrected P-values. Default=0.05.

    cluster_threshold : float, optional
        The threshold to be used to declare clusters as significant. Default=0.001.
    """

    default_threshold_uncorrected_pvalue = 0.001
    default_threshold_corrected_pvalue = 0.05
    default_cluster_threshold = 0.001

    def __init__(
        self,
        design: str,
        df: pd.DataFrame,
        feature_label: str,
        contrast: str,
        fwhm: Optional[int] = 20,
        threshold_uncorrected_pvalue: Optional[float] = None,
        threshold_corrected_pvalue: Optional[float] = None,
        cluster_threshold: Optional[float] = None,
    ):
        self._two_tailed: bool = False
        self._correction = ["fdr", "rft"]
        self.feature_label: str = feature_label
        self.fwhm: Optional[int] = fwhm
        self.threshold_uncorrected_pvalue: float = (
            threshold_uncorrected_pvalue or self.default_threshold_uncorrected_pvalue
        )
        self.threshold_corrected_pvalue: float = (
            threshold_corrected_pvalue or self.default_threshold_corrected_pvalue
        )
        self.cluster_threshold: float = (
            cluster_threshold or self.default_cluster_threshold
        )
        self.results_ = None
        self.slm_models_ = None
        self.filenames: dict = {}
        self.model: FixedEffect = build_model(design, df)
        self.contrasts: List[Contrast] = self._build_contrasts(contrast, df)

    @abc.abstractmethod
    def _build_contrasts(
        self, contrast: str, subjects_df: pd.DataFrame
    ) -> List[Contrast]:
        raise NotImplementedError

    @property
    def contrast_names(self) -> List[str]:
        if self.contrasts is not None:
            return [contrast.name for contrast in self.contrasts]
        return list()

    def get_contrast_by_name(self, contrast_name: str) -> Contrast:
        if self.contrasts:
            contrast = [c for c in self.contrasts if c.name == contrast_name]
            if len(contrast) == 1:
                return contrast[0]
        raise ValueError(f"Unknown contrast {contrast_name}")

    @property
    def results(self):
        if self._is_fitted():
            return self.results_

    def get_output_filename(self, contrast_name: str) -> str:
        """Returns the output file name root for the provided contrast.

        .. note::
            This method needs to be implemented in subclasses.

        Parameters
        ----------
        contrast_name : str
            Contrast for which to get the output filename.
        """
        contrast = self.get_contrast_by_name(contrast_name)
        return self._get_output_filename(contrast)

    @abc.abstractmethod
    def _get_output_filename(self, contrast: Contrast) -> str:
        raise NotImplementedError

    def _is_fitted(self) -> bool:
        return self.results_ is not None

    def fit(
        self, data: np.ndarray, surface: Dict, mask: Optional[np.ndarray] = None
    ) -> None:
        """Fit the GLM model instance.

        Parameters
        ----------
        data : np.ndarray
            The data on which to fit the GLM model.

        surface : dict
            The Brainstat surface on which to fit the GLM model.

        mask : np.ndarray, optional
            The mask to be used to mask the data. Default=None.
        """
        if mask is None:
            mask = data[0, :] > 0
        self.results_ = dict()
        self.slm_models_ = dict()
        for contrast in self.contrasts:
            slm_model = SLM(
                self.model,
                contrast=contrast.built_contrast,
                surf=surface,
                mask=mask,
                two_tailed=self._two_tailed,
                correction=self._correction,
                cluster_threshold=self.cluster_threshold,
            )
            cprint(
                msg=f"Fitting the GLM model with contrast {contrast.name}...",
                lvl="info",
            )
            slm_model.fit(data)
            print_clusters(slm_model, self.threshold_corrected_pvalue)
            self.results_[contrast.name] = StatisticsResults.from_slm_model(
                slm_model,
                mask,
                self.threshold_uncorrected_pvalue,
                self.threshold_corrected_pvalue,
            )
            self.slm_models_[contrast.name] = slm_model

    def save_results(self, output_dir: Path, method: Union[str, List[str]]) -> None:
        """Save results to the provided output directory.

        Parameters
        ----------
        output_dir : PathLike
            The output directory in which to write the results.

        method : str or List[str]
            The method(s) to write the results.
        """
        if not self._is_fitted():
            raise ValueError(
                "GLM model needs to be fitted before accessing the results."
            )
        if isinstance(method, str):
            method = [method]
        for contrast_name, result in self.results_.items():
            result_serializer = StatisticsResultsSerializer(
                output_dir / self.get_output_filename(contrast_name)
            )
            for meth in method:
                result_serializer.save(result, meth)

    def plot_results(
        self,
        output_dir: Path,
        method: Union[str, List[str]],
        mesh: Mesh,
    ) -> None:
        """Plot results to the provided directory.

        Parameters
        ----------
        output_dir : PathLike
            The output directory in which to write the plot files.

        method : str or List[str]
            The method(s) to make the plots.

        mesh : nilearn.surface.Mesh
            The mesh on which to plot the result data.
        """
        if not self._is_fitted():
            raise ValueError(
                "GLM model needs to be fitted before accessing the results."
            )
        if isinstance(method, str):
            method = [method]
        for contrast_name, result in self.results_.items():
            plotter = StatisticsResultsPlotter(
                output_dir / self.get_output_filename(contrast_name), mesh
            )
            for meth in method:
                plotter.plot(result, meth)



"""Module for converting FDG PET of ADNI."""

from enum import Enum
from functools import partial
from os import PathLike
from pathlib import Path
from typing import Iterable, List, Optional, Set, Tuple

import pandas as pd

from clinica.iotools.converters.adni_to_bids.adni_utils import ADNIModalityConverter

__all__ = [
    "convert_fdg_pet",
    "convert_fdg_pet_uniform",
]


class ADNIPreprocessingStep(Enum):
    """ADNI preprocessing steps."""

    STEP0 = "ADNI Brain PET: Raw FDG"
    STEP1 = "Co-registered Dynamic"
    STEP2 = "Co-registered, Averaged"
    STEP3 = "Coreg, Avg, Standardized Image and Voxel Size"
    STEP4_8MM = "Coreg, Avg, Std Img and Vox Siz, Uniform Resolution"
    STEP4_6MM = "Coreg, Avg, Std Img and Vox Siz, Uniform 6mm Res"

    @classmethod
    def from_step_value(cls, step_value: int):
        """Accept step specification in raw integer (0, 1, ..., 5)."""
        error_msg = (
            f"Step value {step_value} is not a valid ADNI preprocessing step value."
            f"Valid values are {list(ADNIPreprocessingStep)}."
        )
        try:
            step_value = int(step_value)
        except Exception:
            raise ValueError(error_msg)
        if 0 <= step_value <= 5:
            if step_value == 4:
                return cls.STEP4_8MM
            if step_value == 5:
                return cls.STEP4_6MM
            return cls[f"STEP{step_value}"]
        raise ValueError(error_msg)


def _convert_fdg_pet(
    source_dir: Path,
    csv_dir: Path,
    destination_dir: Path,
    conversion_dir: Path,
    preprocessing_step: ADNIPreprocessingStep,
    subjects: Iterable[str],
    mod_to_update: bool = False,
    n_procs: int = 1,
):
    """Convert FDG PET images of ADNI into BIDS format.

    Parameters
    ----------
    source_dir : PathLike
        Path to the ADNI directory.

    csv_dir : PathLike
        Path to the clinical data directory.

    destination_dir : PathLike
        Path to the destination BIDS directory.

    conversion_dir : PathLike
        Path to the TSV files including the paths to original images.

    preprocessing_step : ADNIPreprocessingStep
        ADNI processing step.

    subjects : List of str, optional
        List of subjects.

    mod_to_update : bool
        If True, pre-existing images in the BIDS directory
        will be erased and extracted again.

    n_procs : int, optional
        The requested number of processes.
        If specified, it should be between 1 and the number of available CPUs.
        Default=1.
    """
    from clinica.iotools.converters.adni_to_bids.adni_utils import (
        load_clinical_csv,
        paths_to_bids,
    )
    from clinica.utils.stream import cprint

    cprint(
        "Calculating paths of FDG PET images. "
        f"Output will be stored in {conversion_dir}."
    )
    images = _compute_fdg_pet_paths(
        source_dir, csv_dir, subjects, conversion_dir, preprocessing_step
    )

    cprint("Paths of FDG PET images found. Exporting images into BIDS ...")
    modality = _get_modality_from_adni_preprocessing_step(preprocessing_step)
    paths_to_bids(
        images, destination_dir, modality, mod_to_update=mod_to_update, n_procs=n_procs
    )
    cprint(msg="FDG PET conversion done.", lvl="debug")


def _get_modality_from_adni_preprocessing_step(
    step: ADNIPreprocessingStep,
) -> ADNIModalityConverter:
    if step == ADNIPreprocessingStep.STEP2:
        return ADNIModalityConverter.PET_FDG
    if step == ADNIPreprocessingStep.STEP4_8MM:
        return ADNIModalityConverter.PET_FDG_UNIFORM
    raise ValueError(
        f"The ADNI preprocessing step {step} is not (yet) supported by the converter."
        f"The converter only supports {ADNIPreprocessingStep.STEP2} and "
        f"{ADNIPreprocessingStep.STEP4_8MM} for now."
    )


convert_fdg_pet = partial(
    _convert_fdg_pet, preprocessing_step=ADNIPreprocessingStep.STEP2
)
convert_fdg_pet_uniform = partial(
    _convert_fdg_pet, preprocessing_step=ADNIPreprocessingStep.STEP4_8MM
)


def _compute_fdg_pet_paths(
    source_dir: Path,
    csv_dir: Path,
    subjects: Iterable[str],
    conversion_dir: Path,
    preprocessing_step: ADNIPreprocessingStep,
) -> pd.DataFrame:
    """Compute the paths to the FDG PET images and store them in a TSV file.

    Parameters
    ----------
    source_dir : PathLike
        Path to the ADNI directory.

    csv_dir : PathLike
        Path to the clinical data directory. It must contain the following
        CSV files:
            - PETQC.csv
            - PETC3.csv
            - PET_META_LIST.csv

    subjects : list of str
        List of subjects.

    conversion_dir : PathLike
        Path to the TSV files including the paths to original images.

    preprocessing_step : PreprocessingStep
        ADNI processing step, is an int between 0 and 5.

    Returns
    -------
    images : pd.DataFrame
        DataFrame with all the paths to the PET images that will be converted into BIDS.
    """
    from clinica.utils.pet import Tracer

    from ._image_path_utils import find_image_path

    pet_fdg_df = _get_pet_fdg_df(csv_dir, subjects, preprocessing_step)
    images = find_image_path(pet_fdg_df, source_dir, "FDG")
    images.to_csv(
        conversion_dir / f"{Tracer.FDG.value}_pet_paths.tsv",
        sep="\t",
        index=False,
    )

    return images


def _get_pet_fdg_df(
    csv_dir: Path, subjects: Iterable[str], preprocessing_step: ADNIPreprocessingStep
) -> pd.DataFrame:
    """Build a DataFrame for the PET FDG images for the provided list of subjects."""
    import pandas as pd

    dfs = []
    for subject in subjects:
        dfs.extend(
            _get_images_pet_for_subject(
                subject,
                _get_csv_data(Path(csv_dir)),
                preprocessing_step,
            )
        )
    if len(dfs) == 0:
        return pd.DataFrame(columns=_get_pet_fdg_columns())
    pet_fdg_df = pd.concat(dfs, ignore_index=True)
    return _remove_known_conversion_errors(pet_fdg_df)


def _get_pet_fdg_columns() -> List[str]:
    return [
        "Phase",
        "Subject_ID",
        "VISCODE",
        "Visit",
        "Sequence",
        "Scan_Date",
        "Study_ID",
        "Series_ID",
        "Image_ID",
        "Original",
    ]


def _get_csv_data(csv_dir: Path) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame]:
    """Load needed data from .csv files in csv_dir folder."""
    return (
        _get_pet_qc_df(csv_dir),
        _get_qc_adni_3_df(csv_dir),
        _get_meta_list_df(csv_dir),
    )


def _load_df_with_column_check(
    csv_dir: Path, filename: str, required_columns: Set[str]
) -> pd.DataFrame:
    """Load the requested CSV file in a dataframe and check that the requested columns are present."""
    from clinica.iotools.converters.adni_to_bids.adni_utils import load_clinical_csv

    df = load_clinical_csv(csv_dir, filename)
    if not required_columns.issubset(set(df.columns)):
        raise ValueError(
            f"Missing column(s) from {filename} file."
            f"Required columns for this file are {required_columns}."
        )
    return df


_get_pet_qc_df = partial(
    _load_df_with_column_check,
    filename="PETQC",
    required_columns={"PASS", "RID"},
)
_get_qc_adni_3_df = partial(
    _load_df_with_column_check,
    filename="PETC3",
    required_columns={"SCANQLTY", "RID", "SCANDATE"},
)
_get_meta_list_df = partial(
    _load_df_with_column_check,
    filename="PET_META_LIST",
    required_columns={"Subject"},
)


def _get_images_pet_for_subject(
    subject: str,
    csv_data: Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame],
    preprocessing_step: ADNIPreprocessingStep,
) -> List[pd.DataFrame]:
    """Filter the PET images' QC dataframes for the given subject."""
    from ._pet_utils import get_images_pet

    pet_qc_df, pet_qc_adni_3_df, pet_meta_list_df = csv_data
    subject_pet_metadata = pet_meta_list_df[pet_meta_list_df["Subject"] == subject]

    if subject_pet_metadata.empty:
        return []

    subject_pet_qc = _build_pet_qc_all_studies_for_subject(
        subject, pet_qc_df, pet_qc_adni_3_df
    )

    return get_images_pet(
        subject,
        subject_pet_qc,
        subject_pet_metadata,
        _get_pet_fdg_columns(),
        "FDG-PET",
        [preprocessing_step.value],
    )


def _build_pet_qc_all_studies_for_subject(
    subject: str,
    qc_df: pd.DataFrame,
    qc_adni_3_df: pd.DataFrame,
) -> pd.DataFrame:
    """Filter QC for FDG PET images and for the given subject, for studies ADNI 1, GO, 2, and 3.
    Merge everything in a single dataframe.
    """
    subject_rid = _convert_subject_to_rid(subject)
    qc_1go2 = qc_df[(qc_df.PASS == 1) & (qc_df.RID == subject_rid)]
    qc_3 = qc_adni_3_df[
        (qc_adni_3_df.SCANQLTY == 1) & (qc_adni_3_df.RID == subject_rid)
    ]
    qc_3.insert(0, "EXAMDATE", qc_3.SCANDATE.to_list())
    return pd.concat([qc_1go2, qc_3], axis=0, ignore_index=True, sort=False)


def _convert_subject_to_rid(subject: str) -> int:
    """Get the QC RID from the subject string identifier.
    Examples
    --------
    >>> _convert_subject_to_rid("123_S_4567")
    4567
    """

    from re import search

    match = search(r"\d{3}_S_(\d{4})", subject)

    if match:
        return int(match.group(1))
    else:
        raise ValueError(
            f"Cannot convert the subject '{subject}' identifier into a RID "
            "for PET QC filtering. The expected format for the subject is XXX_S_XXXX."
        )


def _remove_known_conversion_errors(df: pd.DataFrame) -> pd.DataFrame:
    """Remove known exceptions from images to convert."""
    if df.empty:
        return df
    error_ind = df.index[df.apply(lambda x: _is_visit_a_conversion_error(x), axis=1)]
    df.drop(error_ind, inplace=True)
    return df


def _is_visit_a_conversion_error(row: pd.Series) -> bool:
    """Return whether a given visit row is among the known exceptions or not."""
    conversion_errors = [  # NONAME.nii
        ("031_S_0294", "bl"),
        ("037_S_1421", "m36"),
        ("037_S_1078", "m36"),
        # Empty folders
        ("941_S_1195", "m48"),
        ("005_S_0223", "m12"),
    ]
    return (row.Subject_ID, row.VISCODE) in conversion_errors



# Use hash instead of parameters for iterables folder names
# Otherwise path will be too long and generate OSError
from pathlib import Path
from typing import List, Optional

from nipype import config

from clinica.pipelines.engine import Pipeline
from clinica.utils.check_dependency import ThirdPartySoftware
from clinica.utils.stream import log_and_warn

cfg = dict(execution={"parameterize_dirs": False})
config.update_config(cfg)


class AnatLinear(Pipeline):
    """Anat Linear - Affine registration of anat (t1w or flair) images to standard space.

    This preprocessing pipeline includes globally three steps:
    1) Bias correction with N4 algorithm from ANTs.
    2) Linear registration to MNI152NLin2009cSym template with
       RegistrationSynQuick from ANTs.
    3) Crop the background (in order to save computational power).

    Returns:
        A clinica pipeline object containing the  AnatLinear pipeline.
    """

    def __init__(
        self,
        bids_directory: Optional[str] = None,
        caps_directory: Optional[str] = None,
        tsv_file: Optional[str] = None,
        overwrite_caps: Optional[bool] = False,
        base_dir: Optional[str] = None,
        parameters: Optional[dict] = None,
        name: Optional[str] = None,
        ignore_dependencies: Optional[List[str]] = None,
        use_antspy: bool = False,
        caps_name: Optional[str] = None,
    ):
        self.use_antspy = use_antspy
        if self.use_antspy:
            if ignore_dependencies:
                ignore_dependencies.append(ThirdPartySoftware.ANTS)
            else:
                ignore_dependencies = [ThirdPartySoftware.ANTS]
            log_and_warn(
                (
                    "The AnatLinear pipeline has been configured to use ANTsPy instead of ANTs.\n"
                    "This means that no installation of ANTs is required, but the antspyx Python "
                    "package must be installed in your environment.\nThis functionality has been "
                    "introduced in Clinica 0.9.0 and is considered experimental.\n"
                    "Please report any issue or unexpected results to the Clinica developer team."
                ),
                UserWarning,
            )
        super().__init__(
            bids_directory=bids_directory,
            caps_directory=caps_directory,
            tsv_file=tsv_file,
            overwrite_caps=overwrite_caps,
            base_dir=base_dir,
            parameters=parameters,
            ignore_dependencies=ignore_dependencies,
            name=name,
            caps_name=caps_name,
        )

    @staticmethod
    def get_processed_images(
        caps_directory: Path, subjects: List[str], sessions: List[str]
    ) -> List[str]:
        from clinica.utils.filemanip import extract_image_ids
        from clinica.utils.input_files import T1W_LINEAR_CROPPED
        from clinica.utils.inputs import clinica_file_reader

        image_ids: List[str] = []
        if caps_directory.is_dir():
            cropped_files, _ = clinica_file_reader(
                subjects, sessions, caps_directory, T1W_LINEAR_CROPPED, False
            )
            image_ids = extract_image_ids(cropped_files)
        return image_ids

    def _check_custom_dependencies(self) -> None:
        """Check dependencies that can not be listed in the `info.json` file."""
        pass

    def _check_pipeline_parameters(self) -> None:
        """Check pipeline parameters."""
        pass

    def get_input_fields(self) -> List[str]:
        """Specify the list of possible inputs of this pipeline.

        Returns
        -------
        list of str :
            A list of (string) input fields name.
        """
        return ["anat"]

    def get_output_fields(self) -> List[str]:
        """Specify the list of possible outputs of this pipeline.

        Returns
        -------
        list of str:
            A list of (string) output fields name.
        """
        return ["image_id"]

    def _build_input_node(self):
        """Build and connect an input node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe

        from clinica.utils.exceptions import ClinicaBIDSError, ClinicaException
        from clinica.utils.filemanip import extract_subjects_sessions_from_filename
        from clinica.utils.image import get_mni_template
        from clinica.utils.input_files import T1W_NII, Flair_T2W_NII
        from clinica.utils.inputs import clinica_file_reader
        from clinica.utils.stream import cprint
        from clinica.utils.ux import print_images_to_process

        self.ref_template = get_mni_template(
            "t1" if self.name == "t1-linear" else "flair"
        )

        # Display image(s) already present in CAPS folder
        # ===============================================
        processed_ids = self.get_processed_images(
            self.caps_directory, self.subjects, self.sessions
        )
        if len(processed_ids) > 0:
            cprint(
                msg=f"Clinica found {len(processed_ids)} image(s) already processed in CAPS directory:",
                lvl="warning",
            )
            for image_id in processed_ids:
                cprint(msg=f"{image_id.replace('_', ' | ')}", lvl="warning")
            cprint(msg=f"Image(s) will be ignored by Clinica.", lvl="warning")
            input_ids = [
                f"{p_id}_{s_id}" for p_id, s_id in zip(self.subjects, self.sessions)
            ]
            to_process_ids = list(set(input_ids) - set(processed_ids))
            self.subjects, self.sessions = extract_subjects_sessions_from_filename(
                to_process_ids
            )

        # Inputs from anat/ folder
        # ========================
        # anat image file:
        try:
            file = T1W_NII if self.name == "t1-linear" else Flair_T2W_NII
            anat_files, _ = clinica_file_reader(
                self.subjects, self.sessions, self.bids_directory, file
            )
        except ClinicaException as e:
            err = (
                "Clinica faced error(s) while trying to read files in your BIDS directory.\n"
                + str(e)
            )
            raise ClinicaBIDSError(err)

        if len(self.subjects):
            print_images_to_process(self.subjects, self.sessions)
            cprint("The pipeline will last approximately 6 minutes per image.")

        read_node = npe.Node(
            name="ReadingFiles",
            iterables=[
                ("anat", anat_files),
            ],
            synchronize=True,
            interface=nutil.IdentityInterface(fields=self.get_input_fields()),
        )
        self.connect(
            [
                (read_node, self.input_node, [("anat", "anat")]),
            ]
        )

    def _build_output_node(self):
        """Build and connect an output node to the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces.io import DataSink

        from clinica.utils.nipype import container_from_filename, fix_join

        if self.name == "flair-linear":
            from .anat_linear_utils import (
                get_substitutions_datasink_flair as get_substitutions,
            )
        else:
            from .anat_linear_utils import (
                get_substitutions_datasink_t1_linear as get_substitutions,
            )

        # Writing node
        write_node = npe.Node(name="WriteCaps", interface=DataSink())
        write_node.inputs.base_directory = str(self.caps_directory)
        write_node.inputs.parameterization = False

        # Other nodes
        # =====================================
        # Get substitutions to rename files
        get_ids = npe.Node(
            interface=nutil.Function(
                input_names=["bids_image_id"],
                output_names=["substitutions"],
                function=get_substitutions,
            ),
            name="GetIDs",
        )
        # Find container path from t1w filename
        container_path = npe.Node(
            nutil.Function(
                input_names=["bids_or_caps_filename"],
                output_names=["container"],
                function=container_from_filename,
            ),
            name="ContainerPath",
        )
        self.connect(
            [
                (self.input_node, container_path, [("anat", "bids_or_caps_filename")]),
                (self.output_node, get_ids, [("image_id", "bids_image_id")]),
                (
                    container_path,
                    write_node,
                    [
                        (
                            ("container", fix_join, self.name.replace("-", "_")),
                            "container",
                        )
                    ],
                ),
                (get_ids, write_node, [("substitutions", "substitutions")]),
                (self.output_node, write_node, [("image_id", "@image_id")]),
                (self.output_node, write_node, [("outfile_reg", "@outfile_reg")]),
                (self.output_node, write_node, [("affine_mat", "@affine_mat")]),
            ]
        )

        if not (self.parameters.get("uncropped_image")):
            self.connect(
                [
                    (self.output_node, write_node, [("outfile_crop", "@outfile_crop")]),
                ]
            )

    def _build_core_nodes(self):
        """Build and connect the core nodes of the pipeline."""
        import nipype.interfaces.utility as nutil
        import nipype.pipeline.engine as npe
        from nipype.interfaces import ants

        from clinica.pipelines.t1_linear.tasks import (
            run_ants_registration_task,
            run_n4biasfieldcorrection_task,
        )
        from clinica.pipelines.tasks import crop_nifti_task, get_filename_no_ext_task

        from .anat_linear_utils import print_end_pipeline

        image_id_node = npe.Node(
            interface=nutil.Function(
                input_names=["filename"],
                output_names=["image_id"],
                function=get_filename_no_ext_task,
            ),
            name="ImageID",
        )

        # 1. N4biascorrection by ANTS. It uses nipype interface.
        n4biascorrection = npe.Node(
            name="n4biascorrection",
            interface=(
                nutil.Function(
                    function=run_n4biasfieldcorrection_task,
                    input_names=[
                        "input_image",
                        "bspline_fitting_distance",
                        "output_prefix",
                        "output_dir",
                        "save_bias",
                        "verbose",
                    ],
                    output_names=["output_image"],
                )
                if self.use_antspy
                else ants.N4BiasFieldCorrection(dimension=3)
            ),
        )
        n4biascorrection.inputs.save_bias = True
        if self.use_antspy:
            n4biascorrection.inputs.output_dir = str(self.base_dir)
            n4biascorrection.inputs.verbose = True
        if self.name == "t1-linear":
            n4biascorrection.inputs.bspline_fitting_distance = 600
        else:
            n4biascorrection.inputs.bspline_fitting_distance = 100

        # 2. `RegistrationSynQuick` by *ANTS*. It uses nipype interface.
        ants_registration_node = npe.Node(
            name="antsRegistrationSynQuick",
            interface=(
                nutil.Function(
                    function=run_ants_registration_task,
                    input_names=[
                        "fixed_image",
                        "moving_image",
                        "random_seed",
                        "output_prefix",
                        "output_dir",
                    ],
                    output_names=["warped_image", "out_matrix"],
                )
                if self.use_antspy
                else ants.RegistrationSynQuick()
            ),
        )
        ants_registration_node.inputs.fixed_image = self.ref_template
        if not self.use_antspy:
            ants_registration_node.inputs.transform_type = "a"
            ants_registration_node.inputs.dimension = 3

        random_seed = self.parameters.get("random_seed", None)
        ants_registration_node.inputs.random_seed = random_seed or 0

        # 3. Crop image (using nifti). It uses custom interface, from utils file

        cropnifti = npe.Node(
            name="cropnifti",
            interface=nutil.Function(
                function=crop_nifti_task,
                input_names=["input_image", "output_path"],
                output_names=["output_img"],
            ),
        )
        cropnifti.inputs.output_path = self.base_dir

        # 4. Print end message
        print_end_message = npe.Node(
            interface=nutil.Function(
                input_names=["anat", "final_file"], function=print_end_pipeline
            ),
            name="WriteEndMessage",
        )
        self.connect(
            [
                (self.input_node, image_id_node, [("anat", "filename")]),
                (self.input_node, n4biascorrection, [("anat", "input_image")]),
                (
                    n4biascorrection,
                    ants_registration_node,
                    [("output_image", "moving_image")],
                ),
                (
                    image_id_node,
                    ants_registration_node,
                    [("image_id", "output_prefix")],
                ),
                # Connect to DataSink
                (image_id_node, self.output_node, [("image_id", "image_id")]),
                (
                    ants_registration_node,
                    self.output_node,
                    [("out_matrix", "affine_mat")],
                ),
                (
                    ants_registration_node,
                    self.output_node,
                    [("warped_image", "outfile_reg")],
                ),
                (self.input_node, print_end_message, [("anat", "anat")]),
            ]
        )
        if self.use_antspy:
            self.connect(
                [
                    (
                        image_id_node,
                        n4biascorrection,
                        [("image_id", "output_prefix")],
                    ),
                ]
            )
        if not (self.parameters.get("uncropped_image")):
            self.connect(
                [
                    (
                        ants_registration_node,
                        cropnifti,
                        [("warped_image", "input_image")],
                    ),
                    (cropnifti, self.output_node, [("output_img", "outfile_crop")]),
                    (cropnifti, print_end_message, [("output_img", "final_file")]),
                ]
            )
        else:
            self.connect(
                [
                    (
                        ants_registration_node,
                        print_end_message,
                        [("warped_image", "final_file")],
                    ),
                ]
            )



"""Methods to find information in the different pipelines of Clinica."""
from enum import Enum
from functools import partial, reduce
from pathlib import Path
from typing import Callable, Iterable, List, Optional, Tuple, Union

import pandas as pd

__all__ = [
    "PipelineNameForMetricExtraction",
    "pipeline_metric_extractor_factory",
]


class PipelineNameForMetricExtraction(str, Enum):
    """Pipelines for which a metric extractor has been implemented."""

    T1_VOLUME = "t1-volume"
    PET_VOLUME = "pet-volume"
    T1_FREESURFER = "t1-freesurfer"
    T1_FREESURFER_LONGI = "t1-freesurfer-longitudinal"
    DWI_DTI = "dwi-dti"


def _extract_metrics_from_pipeline(
    caps_dir: Path,
    df: pd.DataFrame,
    metrics: Iterable[str],
    pipeline: PipelineNameForMetricExtraction,
    atlas_selection: Optional[Iterable[str]] = None,
    group_selection: Optional[Iterable[str]] = None,
    pvc_restriction: Optional[bool] = None,
    tracers_selection: Optional[Iterable[str]] = None,
) -> Tuple[pd.DataFrame, pd.DataFrame]:
    """Extract and merge the data of the provided pipeline into the
    merged dataframe already containing the BIDS information.

    Parameters
    ----------
    caps_dir : Path
        The path to the CAPS directory.

    df : pd.DataFrame
        DataFrame containing the BIDS information.
        New data will be merged into this dataframe.

    metrics : list of str
        List of metrics to extract from the pipeline's caps folder.

    pipeline : PipelineNameForMetricExtraction
        Name of the pipeline for which to extract information.

    atlas_selection : list of str, optional
        Allows to choose the atlas to merge.
        If None all atlases are selected.

    group_selection : list of str, optional
        Allows to choose the DARTEL groups to merge.
        If None all groups are selected.

    pvc_restriction : bool, optional
        If True, only the atlases containing the label 'pvc-rbv' will be used.
        If False, the atlases containing the label won't be used.
        If None, all the atlases will be used.

    tracers_selection : list of str, optional
        Allows to choose the PET tracer to merge.
        If None, all tracers available will be used.

    Returns
    -------
    final_df : pd.DataFrame
        DataFrame containing the information from the bids dataset as well
        as the information extracted from the caps pipeline's folder.

    summary_df : pd.DataFrame
        Summary DataFrame generated by function `generate_summary`.
    """
    if df.index.names != ["participant_id", "session_id"]:
        try:
            df.set_index(
                ["participant_id", "session_id"], inplace=True, verify_integrity=True
            )
        except KeyError:
            raise KeyError("Fields `participant_id` and `session_id` are required.")
    group_selection = _check_group_selection(caps_dir, group_selection)
    records = _get_records(
        caps_dir,
        df,
        pipeline,
        metrics,
        group_selection,
        atlas_selection,
        pvc_restriction,
        tracers_selection,
    )
    if records:
        pipeline_df = pd.DataFrame.from_records(
            records, index=["participant_id", "session_id"]
        )
    else:
        pipeline_df = pd.DataFrame.from_records(
            records, columns=["participant_id", "session_id"]
        )
        pipeline_df.set_index(
            ["participant_id", "session_id"], inplace=True, verify_integrity=True
        )
    summary_df = _generate_summary(
        pipeline_df, pipeline, ignore_groups=group_selection == [""]
    )
    final_df = pd.concat([df, pipeline_df], axis=1)
    final_df.reset_index(inplace=True)

    return final_df, summary_df


_extract_metrics_from_dwi_dti = partial(
    _extract_metrics_from_pipeline,
    metrics=["FA_statistics", "MD_statistics", "RD_statistics", "AD_statistics"],
    pipeline=PipelineNameForMetricExtraction.DWI_DTI,
    group_selection=[""],
)


_extract_metrics_from_t1_freesurfer_longitudinal = partial(
    _extract_metrics_from_pipeline,
    metrics=["volume", "thickness", "segmentationVolumes"],
    pipeline=PipelineNameForMetricExtraction.T1_FREESURFER_LONGI,
    group_selection=[""],
)

_extract_metrics_from_t1_freesurfer = partial(
    _extract_metrics_from_pipeline,
    metrics=["thickness", "segmentationVolumes"],
    pipeline=PipelineNameForMetricExtraction.T1_FREESURFER,
    group_selection=[""],
)

_extract_metrics_from_t1_volume = partial(
    _extract_metrics_from_pipeline,
    metrics=["statistics"],
    pipeline=PipelineNameForMetricExtraction.T1_VOLUME,
)

_extract_metrics_from_pet_volume = partial(
    _extract_metrics_from_pipeline,
    metrics=["statistics"],
    pipeline=PipelineNameForMetricExtraction.PET_VOLUME,
)


def pipeline_metric_extractor_factory(
    name: Union[str, PipelineNameForMetricExtraction],
) -> Callable:
    """Factory returning a metric extractor given its name."""
    name = PipelineNameForMetricExtraction(name)
    if name == PipelineNameForMetricExtraction.T1_VOLUME:
        return _extract_metrics_from_t1_volume
    if name == PipelineNameForMetricExtraction.PET_VOLUME:
        return _extract_metrics_from_pet_volume
    if name == PipelineNameForMetricExtraction.T1_FREESURFER:
        return _extract_metrics_from_t1_freesurfer
    if name == PipelineNameForMetricExtraction.T1_FREESURFER_LONGI:
        return _extract_metrics_from_t1_freesurfer_longitudinal
    if name == PipelineNameForMetricExtraction.DWI_DTI:
        return _extract_metrics_from_dwi_dti


def _check_group_selection(
    caps_dir: Path, group_selection: Optional[Iterable[str]] = None
) -> Iterable[str]:
    if group_selection == [""]:
        return group_selection
    if group_selection is None:
        return [f.name for f in (caps_dir / "groups").iterdir()]
    return [f"group-{group}" for group in group_selection]


def _get_records(
    caps_dir: Path,
    df: pd.DataFrame,
    pipeline: PipelineNameForMetricExtraction,
    metrics: Iterable[str],
    group_selection: Iterable[str],
    atlas_selection: Optional[Iterable[str]] = None,
    pvc_restriction: Optional[bool] = None,
    tracers_selection: Optional[Iterable[str]] = None,
) -> List[dict]:
    """Returns a list of dictionaries corresponding to the dataframe rows of the pipeline dataframe."""
    from clinica.utils.stream import cprint

    records = []
    subjects_dir = caps_dir / "subjects"
    for participant_id, session_id in df.index.values:
        if (
            mod_path := _get_modality_path(
                subjects_dir / participant_id / session_id, pipeline
            )
        ) is None or not mod_path.exists():
            cprint(
                f"Could not find a longitudinal dataset for participant {participant_id} {session_id}",
                lvl="warning",
            )
            continue
        for group_path in (
            mod_path / group for group in group_selection if (mod_path / group).exists()
        ):
            records_of_group = [
                _get_single_record(
                    group_path,
                    metric,
                    participant_id,
                    session_id,
                    pipeline,
                    atlas_selection,
                    pvc_restriction,
                    tracers_selection,
                )
                for metric in metrics
            ]
            records.append(reduce(lambda x, y: {**x, **y}, records_of_group))
    return records


def _get_modality_path(
    ses_path: Path, pipeline: PipelineNameForMetricExtraction
) -> Optional[Path]:
    """Returns the path to the modality of interest depending on the pipeline considered."""
    if pipeline == PipelineNameForMetricExtraction.DWI_DTI:
        return ses_path / "dwi" / "dti_based_processing" / "atlas_statistics"
    if pipeline == PipelineNameForMetricExtraction.T1_FREESURFER_LONGI:
        mod_path = ses_path / "t1"
        if len(long_ids := list(mod_path.glob("long*"))) == 0:
            return None
        return (
            mod_path
            / long_ids[0].name
            / "freesurfer_longitudinal"
            / "regional_measures"
        )
    if pipeline == PipelineNameForMetricExtraction.T1_FREESURFER:
        return ses_path / "t1" / "freesurfer_cross_sectional" / "regional_measures"
    if pipeline == PipelineNameForMetricExtraction.T1_VOLUME:
        return ses_path / "t1" / "spm" / "dartel"
    if pipeline == PipelineNameForMetricExtraction.PET_VOLUME:
        return ses_path / "pet" / "preprocessing"


def _get_single_record(
    group_path: Path,
    metric: str,
    participant_id: str,
    session_id: str,
    pipeline: PipelineNameForMetricExtraction,
    atlas_selection: Optional[Iterable[str]] = None,
    pvc_restriction: Optional[bool] = None,
    tracers_selection: Optional[Iterable[str]] = None,
) -> dict:
    """Get a single record (dataframe row) for a given participant, session, group, and metric."""
    atlases = [
        atlas_path
        for atlas_path in _get_atlas_paths(
            group_path, participant_id, session_id, metric
        )
        if not _skip_atlas(
            atlas_path, pipeline, pvc_restriction, tracers_selection, atlas_selection
        )
    ]
    return {
        **{"participant_id": participant_id, "session_id": session_id},
        **reduce(
            lambda a, b: {**a, **b},
            [
                _get_records_for_atlas(atlas, pipeline, metric, group_path.name)
                for atlas in atlases
            ],
        ),
    }


def _get_atlas_paths(
    group_path: Path, participant_id: str, session_id: str, metric: str
) -> List[Path]:
    """Returns a list of paths to atlases within the provided group folder and
    matching the provided participant, session, and metric.
    """
    if (
        len(
            paths := sorted(
                group_path.glob(f"{participant_id}_{session_id}_*{metric}.tsv")
            )
        )
        != 0
    ):
        return paths
    return sorted(
        (group_path / "atlas_statistics").glob(
            f"{participant_id}_{session_id}_*{metric}.tsv"
        )
    )


def _skip_atlas(
    atlas_path: Path,
    pipeline: PipelineNameForMetricExtraction,
    pvc_restriction: Optional[bool] = None,
    tracers_selection: Optional[Iterable[str]] = None,
    atlas_selection: Optional[Iterable[str]] = None,
) -> bool:
    """Returns whether the atlas provided through its path should be skipped or not."""
    if not atlas_path.exists():
        return True
    if _skip_atlas_based_on_pipeline(
        atlas_path, pipeline, pvc_restriction, tracers_selection
    ):
        return True
    return _skip_atlas_based_on_selection(atlas_path, pipeline, atlas_selection)


def _skip_atlas_based_on_pipeline(
    atlas_path: Path,
    pipeline: PipelineNameForMetricExtraction,
    pvc_restriction: Optional[bool] = None,
    tracers_selection: Optional[Iterable[str]] = None,
) -> bool:
    """Returns True if the atlas provided through its path should be skipped based on the pipeline name."""
    if pipeline == PipelineNameForMetricExtraction.T1_FREESURFER_LONGI:
        return "-wm_" in str(atlas_path) or "-ba_" in str(atlas_path.stem)
    if pipeline in (
        PipelineNameForMetricExtraction.T1_VOLUME,
        PipelineNameForMetricExtraction.PET_VOLUME,
    ):
        skip = []
        if pvc_restriction is not None:
            if pvc_restriction:
                skip.append("pvc-rbv" not in str(atlas_path))
            else:
                skip.append("pvc-rbv" in str(atlas_path))
        if tracers_selection:
            skip.append(
                all([tracer not in str(atlas_path) for tracer in tracers_selection])
            )
        return any(skip)
    return False


def _skip_atlas_based_on_selection(
    atlas_path: Path,
    pipeline: PipelineNameForMetricExtraction,
    atlas_selection: Optional[Iterable[str]] = None,
) -> bool:
    """Returns True if the atlas provided through its path should be skipped based on the user-provided selection."""
    return (
        atlas_selection is not None
        and _get_atlas_name(atlas_path, pipeline) not in atlas_selection
    )


def _get_records_for_atlas(
    atlas_path: Path,
    pipeline: PipelineNameForMetricExtraction,
    metric: str,
    group: str,
) -> dict:
    atlas_df = pd.read_csv(atlas_path, sep="\t")
    label_list = _get_label_list(atlas_path, metric, pipeline, group)
    key = "label_value" if "freesurfer" in pipeline.value else "mean_scalar"
    values = atlas_df[key].to_numpy()
    return {label: value for label, value in zip(label_list, values)}


def _get_label_list(
    atlas_path: Path, metric: str, pipeline: PipelineNameForMetricExtraction, group: str
) -> List[str]:
    """Returns the list of labels to use in the session df depending on the
    pipeline, the atlas, and the metric considered.
    """
    from clinica.iotools.converter_utils import replace_sequence_chars

    atlas_df = pd.read_csv(atlas_path, sep="\t")
    atlas_name = _get_atlas_name(atlas_path, pipeline)
    if pipeline == PipelineNameForMetricExtraction.T1_FREESURFER:
        return [
            f"t1-freesurfer_{atlas_name}_ROI-{replace_sequence_chars(roi_name)}_{'volume' if metric == 'segmentationVolumes' else metric}"
            for roi_name in atlas_df.label_name.values
        ]
    if pipeline in (
        PipelineNameForMetricExtraction.T1_VOLUME,
        PipelineNameForMetricExtraction.PET_VOLUME,
    ):
        additional_desc = ""
        if "trc" in str(atlas_path):
            tracer = str(atlas_path).split("_trc-")[1].split("_")[0]
            additional_desc += f"_trc-{tracer}"
        if "pvc-rbv" in str(atlas_path):
            additional_desc += f"_pvc-rbv"
        return [
            f"{pipeline.value}_{group}_{atlas_name}{additional_desc}_ROI-{replace_sequence_chars(roi_name)}_intensity"
            for roi_name in atlas_df.label_name.values
        ]
    if pipeline == PipelineNameForMetricExtraction.DWI_DTI:
        prefix = "dwi-dti_"
        metric = metric.rstrip("_statistics")
    else:
        prefix = "t1-fs-long_"
    return [prefix + metric + atlas_name + "_" + x for x in atlas_df.label_name.values]


def _get_atlas_name(atlas_path: Path, pipeline: PipelineNameForMetricExtraction) -> str:
    """Return the atlas name, inferred from the atlas_path, for a given pipeline."""
    if pipeline == PipelineNameForMetricExtraction.DWI_DTI:
        return _infer_atlas_name("_dwi_space-", atlas_path)
    if pipeline in (
        PipelineNameForMetricExtraction.T1_FREESURFER_LONGI,
        PipelineNameForMetricExtraction.T1_FREESURFER,
    ):
        return _infer_atlas_name("_parcellation-", atlas_path)
    if pipeline in (
        PipelineNameForMetricExtraction.T1_VOLUME,
        PipelineNameForMetricExtraction.PET_VOLUME,
    ):
        return _infer_atlas_name("_space-", atlas_path)


def _infer_atlas_name(splitter: str, atlas_path: Path) -> str:
    try:
        assert splitter in atlas_path.stem
        return f"atlas-{atlas_path.stem.split(splitter)[-1].split('_')[0]}"
    except Exception:
        if "segmentationVolumes" in atlas_path.stem:
            return "segmentation-volumes"
        else:
            raise ValueError(f"Unable to infer the atlas name from {atlas_path}.")


def _generate_summary(
    pipeline_df: pd.DataFrame,
    pipeline_name: PipelineNameForMetricExtraction,
    ignore_groups: bool = False,
):
    columns = [
        "pipeline_name",
        "group_id",
        "atlas_id",
        "tracer",
        "pvc",
        "regions_number",
        "first_column_name",
        "last_column_name",
    ]
    summary_df = pd.DataFrame(columns=columns)

    if ignore_groups:
        groups = ["_"]
        atlases = list({column.split("_")[1] for column in pipeline_df.columns.values})
    else:
        groups = list({column.split("_")[1] for column in pipeline_df.columns.values})
        atlases = list({column.split("_")[2] for column in pipeline_df.columns.values})
    pvc_restrictions = list(
        {"pvc-rbv" in column for column in pipeline_df.columns.values}
    )
    tracers = list(
        {
            column.split("_trc-")[1].split("_")[0]
            for column in pipeline_df.columns.values
            if "trc" in column
        }
    )
    if len(tracers) == 0:
        tracers.append("_")
        pvc_restrictions = ["_"]
    groups.sort()
    atlases.sort()

    for group in groups:
        group_id = group.split("-")[-1]
        for atlas in atlases:
            atlas_id = atlas.split("-")[-1]
            for tracer in tracers:
                for pvc_restriction in pvc_restrictions:
                    if pvc_restriction and pvc_restriction != "_":
                        regions = [
                            column
                            for column in pipeline_df.columns.values
                            if group in column
                            and atlas in column
                            and "pvc-rbv" in column
                            and tracer in column
                        ]
                    else:
                        regions = [
                            column
                            for column in pipeline_df.columns.values
                            if group in column
                            and atlas in column
                            and "pvc-rbv" not in column
                            and tracer in column
                        ]

                    if len(regions) > 0:
                        row_df = pd.DataFrame(
                            [
                                [
                                    pipeline_name.value,
                                    group_id,
                                    atlas_id,
                                    tracer,
                                    pvc_restriction,
                                    len(regions),
                                    regions[0],
                                    regions[-1],
                                ]
                            ],
                            columns=columns,
                        )
                        summary_df = pd.concat([summary_df, row_df])

    summary_df = summary_df.replace("_", "n/a")
    return summary_df



from pathlib import Path
from typing import List, Optional, Union

import click

bids_directory = click.argument(
    "bids_directory", type=click.Path(exists=True, resolve_path=True)
)

current_directory = click.argument(
    "current_directory", type=click.Path(exists=True, resolve_path=True)
)


@click.group("iotools")
def cli() -> None:
    """Tools to handle BIDS/CAPS datasets."""
    pass


@cli.command()
@click.argument("dataset", type=click.Path(resolve_path=True))
def describe(dataset: Union[str, Path]):
    """Describe a dataset in BIDS or CAPS format."""
    from .describe import describe as _describe

    _describe(dataset)


@cli.command()
@bids_directory
@click.argument("output_bids_directory", type=click.Path(resolve_path=True))
@click.option(
    "-m",
    "--modality",
    "modalities",
    multiple=True,
    metavar="modalities",
    help="Process selected modalities.",
)
@click.option("--center-all-files", is_flag=True, help="Force processing of all files.")
def center_nifti(
    bids_directory: str,
    output_bids_directory: str,
    modalities: Optional[List[str]] = None,
    center_all_files: bool = False,
) -> None:
    """Center NIfTI files in a BIDS dataset."""
    import sys

    from clinica.utils.exceptions import ClinicaExistingDatasetError

    from .center_nifti import center_nifti as center_nifti_

    try:
        center_nifti_(
            bids_directory,
            output_bids_directory,
            modalities=modalities,
            center_all_files=center_all_files,
            overwrite_existing_files=False,
        )
    except ClinicaExistingDatasetError:
        click.echo(
            "Target BIDS directory is not empty. Existing files may be overwritten."
        )
        if click.confirm("Do you wish to continue?"):
            center_nifti_(
                bids_directory,
                output_bids_directory,
                modalities=modalities,
                center_all_files=center_all_files,
                overwrite_existing_files=True,
            )
        else:
            click.echo("Clinica will now exit...")
            sys.exit(0)


@cli.command()
@bids_directory
@click.argument("output_directory", type=click.Path(writable=True))
@click.option(
    "-op",
    "--output_prefix",
    default="missing_mods",
    show_default=True,
    help="Prefix for the name of the output files.",
)
def check_missing_modalities(
    bids_directory: str,
    output_directory: str,
    output_prefix: str = "missing_mods",
) -> None:
    """Check missing modalities in a BIDS dataset."""
    from clinica.iotools.utils.data_handling import compute_missing_mods
    from clinica.utils.inputs import check_bids_folder

    check_bids_folder(bids_directory)
    compute_missing_mods(bids_directory, output_directory, output_prefix)


@cli.command()
@bids_directory
@click.argument("caps_directory", type=click.Path(exists=True, resolve_path=True))
@click.argument("output_file", type=click.Path(resolve_path=True))
def check_missing_processing(
    bids_directory: str,
    caps_directory: str,
    output_file: str,
) -> None:
    """Check missing processing in a CAPS dataset."""
    from clinica.iotools.utils.data_handling import compute_missing_processing
    from clinica.utils.inputs import check_caps_folder

    check_caps_folder(caps_directory)
    compute_missing_processing(bids_directory, caps_directory, output_file)


@cli.command()
@current_directory
@click.argument("output_tsv", type=click.Path(resolve_path=True))
def create_subjects_visits(current_directory: str, output_tsv: str) -> None:
    """Export participants with their sessions."""
    from os import makedirs
    from os.path import basename, dirname

    from clinica.iotools.utils.data_handling import create_subs_sess_list
    from clinica.utils.inputs import determine_caps_or_bids
    from clinica.utils.stream import cprint

    is_bids = determine_caps_or_bids(current_directory)
    output_directory = dirname(output_tsv)
    makedirs(output_directory, exist_ok=True)
    create_subs_sess_list(
        current_directory, output_directory, basename(output_tsv), is_bids_dir=is_bids
    )
    cprint(f"The TSV file was saved to {output_tsv}.")


@cli.command()
@bids_directory
@click.argument("output_tsv", type=click.Path(resolve_path=True))
@click.option(
    "-caps",
    "--caps_directory",
    type=click.Path(exists=True),
    help="Path to a CAPS dataset.",
)
@click.option(
    "-p",
    "--pipeline",
    "pipelines",
    multiple=True,
    type=click.Choice(
        [
            "t1-freesurfer",
            "t1-volume",
            "pet-volume",
            "t1-freesurfer-longitudinal",
            "dwi-dti",
        ]
    ),
    help="Pipeline to merge to the output TSV file. All pipelines are merged by default.",
)
@click.option(
    "-vas",
    "--volume_atlas_selection",
    multiple=True,
    help="Atlas to merge for t1- and pet-volume. All atlases are merged by default.",
)
@click.option(
    "-fas",
    "--freesurfer_atlas_selection",
    multiple=True,
    help="Atlas to merge for t1-freesurfer. All atlases are merged by default",
)
@click.option(
    "-pvc",
    "--pvc_restriction",
    type=click.IntRange(0, 1),
    help=(
        "Restriction on the label [_pvc-rbv] for pet-volume. Default: Merge all atlases, "
        "0: Merge atlases without the label only, 1: Merge atlases with the label only."
    ),
)
@click.option(
    "-pts",
    "--pet_tracers_selection",
    multiple=True,
    help="PET tracer to merge. All PET tracers are merged by default.",
)
@click.option(
    "-group",
    "--group_selection",
    multiple=True,
    help="Group to merge. All groups are merged by default.",
)
@click.option(
    "-tsv",
    "--subjects_sessions_tsv",
    type=click.Path(exists=True, resolve_path=True),
    help="List of subjects and their sessions in TSV format.",
)
@click.option(
    "--ignore_scan_files",
    is_flag=True,
    help="Ignore scan files. This may accelerate the procedure.",
)
@click.option(
    "--ignore_session_scan_files",
    is_flag=True,
    help="Ignore session files. This may accelerate the procedure.",
)
def merge_tsv(
    bids_directory: str,
    output_tsv: str,
    caps_directory: Optional[str] = None,
    pipelines: Optional[List[str]] = None,
    volume_atlas_selection: Optional[List[str]] = None,
    freesurfer_atlas_selection: Optional[List[str]] = None,
    pvc_restriction: Optional[int] = None,
    pet_tracers_selection: Optional[List[str]] = None,
    group_selection: Optional[List[str]] = None,
    subjects_sessions_tsv: Optional[str] = None,
    ignore_scan_files: bool = False,
    ignore_session_scan_files: bool = False,
) -> None:
    """Merge clinical data into a single TSV file."""
    from .merge_tsv import merge_tsv as merge_tsv_

    merge_tsv_(
        bids_directory,
        output_tsv,
        caps_directory=caps_directory,
        pipelines=pipelines,
        volume_atlas_selection=volume_atlas_selection,
        freesurfer_atlas_selection=freesurfer_atlas_selection,
        pvc_restriction=pvc_restriction,
        pet_tracers_selection=pet_tracers_selection,
        group_selection=group_selection,
        subjects_sessions_tsv=subjects_sessions_tsv,
        ignore_scan_files=ignore_scan_files,
        ignore_session_scan_files=ignore_session_scan_files,
    )



from os import PathLike
from typing import Optional

import click

from clinica.iotools.converters import cli_param

clinical_data_directory = click.option(
    "-cdd",
    "--clinical-data-dir",
    "clinical_data_directory",
    type=click.Path(exists=True, file_okay=False, resolve_path=True),
    help="Path to the clinical data directory",
)

gif = click.option("-gif", is_flag=True, help="Add values from gif to session.tsv")
clinical_data_tsv = click.option(
    "-cdt",
    "--clinical-data-tsv",
    "clinical_data_tsv",
    type=click.Path(exists=True, file_okay=True, resolve_path=True),
    help="Path to a tsv file containing additional clinical data you want to have in the BIDS output.",
)


@click.command(name="genfi-to-bids")
@cli_param.dataset_directory
@cli_param.bids_directory
@clinical_data_directory
@gif
@clinical_data_tsv
def cli(
    dataset_directory: PathLike,
    bids_directory: PathLike,
    clinical_data_directory: Optional[PathLike] = None,
    clinical_data_tsv: Optional[PathLike] = None,
    gif: Optional[bool] = False,
) -> None:
    """GENFI to BIDS converter.

    Convert the imaging and clinical data of GENFI, located in DATASET_DIRECTORY and
    CLINICAL_DATA_DIRECTORY respectively, to a BIDS dataset in the target BIDS_DIRECTORY.
    """
    from .genfi_to_bids import convert

    convert(
        dataset_directory,
        bids_directory,
        clinical_data_directory,
        gif=gif,
        path_to_clinical_tsv=clinical_data_tsv,
    )


if __name__ == "__main__":
    cli()
